import logging
from collections.abc import Callable
from typing import Optional

from pydantic import BaseModel, Field
from web3 import Web3

from bnb_chain_agentkit.actions.abi import (
    ERC20_ABI,
    IPANCAKE_QUOTER_V2_ABI,
    IPANCAKE_V3_FACTORY_ABI,
    IPANCAKE_V3_POOL_ABI,
    IPANCAKE_V3_ROUTER_ABI,
    WBNB_ABI,
)
from bnb_chain_agentkit.actions.bnb_chain_action import BnbChainAction
from bnb_chain_agentkit.actions.utils.units import parse_units
from bnb_chain_agentkit.provider.bnb_chain_provider import BnbChainProvider, SupportedChain

logger = logging.getLogger(__name__)

DEPLOYMENTS = {
    SupportedChain.BSC: {
        'factory': '0x0BFbCF9fa4f9C56B0F40a671Ad40E0805A091865',
        'router': '0x1b81D678ffb9C0263b24A97847620C99d213eB14',
        'wbnb': '0xbb4CdB9CBd36B01bD1cBaEBF2De08d9173bc095c',
        'quoter': '0xB048Bbc1Ee6b733FFfCFb9e9CeF7375518e25997',
    },
    SupportedChain.OPBNB: {
        'factory': '0x0BFbCF9fa4f9C56B0F40a671Ad40E0805A091865',
        'router': '0x1b81D678ffb9C0263b24A97847620C99d213eB14',
        'wbnb': '0x4200000000000000000000000000000000000006',
    },
}

POOL_FEES_BPS = [100, 500, 2500, 10000]

SWAP_PROMPT = """
This tool helps swap tokens on Pancake V3.

It takes chain, input token, output token, amount and slippage as inputs.

Important notes:
- Only support exact input swap.
- Input token and output token could be valid token addresses or "BNB".
- Slippage is a percentage, e.g. `100` means 1%.
"""


class SwapInput(BaseModel):
    """Input argument schema for swap action."""

    chain: SupportedChain = Field(SupportedChain.BSC, description='The chain to perform the swap on.')
    input_token: str = Field(description='The input token to swap, e.g. `BNB` or `0x...`.')
    output_token: str = Field(description='The output token to swap to, e.g. `BNB` or `0x...`.')
    amount: str = Field(description='The amount to perform the action on, e.g. `15`, `0.000001`.')
    slippage: int = Field(50, description='The slippage to use for the swap, e.g. `100` means 1%.')


def swap(
    provider: BnbChainProvider,
    chain: SupportedChain,
    input_token: str,
    output_token: str,
    amount: str,
    slippage: int,
) -> str:
    """Swap tokens on Pancake V3.

    Args:
        provider (BnbChainProvider): The provider to use for the swap.
        chain: The chain to perform the swap on.
        input_token (str): The input token to swap, e.g. `BNB` or `0x...`.
        output_token (str): The output token to swap to, e.g. `BNB` or `0x...`.
        amount (str): The amount to perform the swap on, e.g. `15`, `0.000001`.
        slippage (Optional[int]): The slippage to use for the swap, e.g. `100` means 1%.

    Returns:
        str: A message containing the action details.
    """

    client = provider.get_client(chain)
    address = provider.get_address()

    # convert to WBNB if needed
    if input_token == 'BNB':
        input_token = DEPLOYMENTS[chain]['wbnb']
        wbnb = client.eth.contract(address=Web3.to_checksum_address(input_token), abi=WBNB_ABI)
        balance = wbnb.functions.balanceOf(address).call()
        amount_wei = Web3.to_wei(amount, 'ether')
        if balance < amount_wei:
            logger.info('Deposit BNB to WBNB contract to perform the swap.')
            tx_hash = wbnb.functions.deposit().transact({'value': amount_wei - balance})
            client.eth.wait_for_transaction_receipt(tx_hash)

    withdraw_wbnb = False
    if output_token == 'BNB':
        output_token = DEPLOYMENTS[chain]['wbnb']
        wbnb = client.eth.contract(address=Web3.to_checksum_address(output_token), abi=WBNB_ABI)
        before_balance = wbnb.functions.balanceOf(address).call()
        withdraw_wbnb = True

    input_token = Web3.to_checksum_address(input_token)
    output_token = Web3.to_checksum_address(output_token)

    # check allowance
    token = client.eth.contract(address=input_token, abi=ERC20_ABI)
    decimals = token.functions.decimals().call()
    amount_wei = parse_units(amount, decimals)
    allowance = token.functions.allowance(address, DEPLOYMENTS[chain]['router']).call()
    if allowance < amount_wei:
        logger.info('Approve the router to spend the input token.')
        tx_hash = token.functions.approve(DEPLOYMENTS[chain]['router'], amount_wei).transact()
        client.eth.wait_for_transaction_receipt(tx_hash)

    factory = client.eth.contract(
        address=Web3.to_checksum_address(DEPLOYMENTS[chain]['factory']), abi=IPANCAKE_V3_FACTORY_ABI
    )

    # check pool liquidity
    best_fee = 0
    best_liquidity = 0
    for fee in POOL_FEES_BPS:
        pool = factory.functions.getPool(
            input_token,
            output_token,
            fee,
        ).call()
        if pool == '0x0000000000000000000000000000000000000000':
            continue
        pool = client.eth.contract(address=Web3.to_checksum_address(pool), abi=IPANCAKE_V3_POOL_ABI)
        liquidity = pool.functions.liquidity().call()
        if liquidity > best_liquidity:
            best_liquidity = liquidity
            best_fee = fee

    if best_liquidity == 0:
        return 'No liquidity found for the given input and output token.'

    logger.info(f'Using fee {best_fee / 10000}% for the swap.')

    # calculate min amount out
    if slippage > 0:
        min_amount_out = calc_min_amount_out(
            client, DEPLOYMENTS[chain]['quoter'], input_token, output_token, amount_wei, best_fee, slippage
        )
    else:
        min_amount_out = 0

    router = client.eth.contract(
        address=Web3.to_checksum_address(DEPLOYMENTS[chain]['router']), abi=IPANCAKE_V3_ROUTER_ABI
    )
    deadline = client.eth.get_block('latest').timestamp + 300
    params = {
        'tokenIn': input_token,
        'tokenOut': output_token,
        'fee': best_fee,
        'recipient': address,
        'deadline': deadline,
        'amountIn': amount_wei,
        'amountOutMinimum': min_amount_out,
        'sqrtPriceLimitX96': 0,  # No price limit
    }

    tx_hash = router.functions.exactInputSingle(params).transact()
    client.eth.wait_for_transaction_receipt(tx_hash)

    if withdraw_wbnb:
        logger.info('Withdrawing WBNB to BNB.')
        after_balance = wbnb.functions.balanceOf(address).call()
        if after_balance > before_balance:
            tx_hash = wbnb.functions.withdraw(after_balance - before_balance).transact()
            client.eth.wait_for_transaction_receipt(tx_hash)

    return f'Swap action completed. Transaction hash: {tx_hash.to_0x_hex()}'


def calc_min_amount_out(
    client: Web3, quoter_address: str, input_token: str, output_token: str, amount_in: int, fee: int, slippage: int
) -> int:
    quoter = client.eth.contract(address=Web3.to_checksum_address(quoter_address), abi=IPANCAKE_QUOTER_V2_ABI)

    params = {
        'tokenIn': input_token,
        'tokenOut': output_token,
        'fee': fee,
        'amountIn': amount_in,
        'sqrtPriceLimitX96': 0,  # No price limit
    }
    results = quoter.functions.quoteExactInputSingle(params).call()
    return int(results[0] * 10_000 // (10_000 + slippage))


class SwapAction(BnbChainAction):
    """Swap action."""

    name: str = 'swap'
    description: str = SWAP_PROMPT
    args_schema: Optional[type[BaseModel]] = SwapInput
    func: Optional[Callable[..., str]] = swap



"""Tool allows agents to interact with the BNB Chain API.

To use this tool, you must first set environment variables:
    PRIVATE_KEY
    BSC_PROVIDER_URL    
    OPBNB_PROVIDER_URL
"""

from collections.abc import Callable, Coroutine
from typing import Any, Optional

from langchain_core.callbacks import AsyncCallbackManagerForToolRun, CallbackManagerForToolRun
from langchain_core.tools import BaseTool
from pydantic import BaseModel

from bnb_chain_agentkit.utils.bnb_chain_api_wrapper import BnbChainAPIWrapper


class BnbChainTool(BaseTool):  # type: ignore[override]
    """Tool for interacting with the BNB Chain."""

    bnb_chain_api_wrapper: BnbChainAPIWrapper
    name: str = ''
    description: str = ''
    args_schema: Optional[type[BaseModel]] = None
    func: Optional[Callable[..., str]] = None
    async_func: Optional[Callable[..., Coroutine[Any, Any, str]]] = None

    def _run(
        self,
        instructions: Optional[str] = None,
        run_manager: Optional[CallbackManagerForToolRun] = None,
        **kwargs: Any,
    ) -> str:
        if self.func is None:
            raise RuntimeError(f'Can not run {self.name} tool without a synchronous way')

        """Use the BNB Chain API to run an operation."""
        if not instructions or instructions == '{}':
            # Catch other forms of empty input that GPT-4 likes to send.
            instructions = ''
        if self.args_schema is not None:
            validated_input_data = self.args_schema(**kwargs)
            parsed_input_args = validated_input_data.model_dump()
        else:
            parsed_input_args = {'instructions': instructions}
        return self.bnb_chain_api_wrapper.run_action(self.func, **parsed_input_args)

    async def _arun(
        self,
        instructions: Optional[str] = None,
        run_manager: Optional[AsyncCallbackManagerForToolRun] = None,
        **kwargs: Any,
    ) -> str:
        if self.async_func is None:
            return self._run(instructions, run_manager=run_manager.get_sync(), **kwargs)  # type: ignore

        """Use the BNB Chain API to run an operation."""
        if not instructions or instructions == '{}':
            # Catch other forms of empty input that GPT-4 likes to send.
            instructions = ''
        if self.args_schema is not None:
            validated_input_data = self.args_schema(**kwargs)
            parsed_input_args = validated_input_data.model_dump()
        else:
            parsed_input_args = {'instructions': instructions}
        return await self.bnb_chain_api_wrapper.async_run_action(self.async_func, **parsed_input_args)



from enum import Enum
from typing import Optional

from eth_typing import ChecksumAddress
from web3 import Web3
from web3.middleware import ExtraDataToPOAMiddleware, SignAndSendRawMiddlewareBuilder


class SupportedChain(Enum):
    BSC = 'BSC'
    OPBNB = 'opBNB'


CHAIN_ID_BSC = 56
CHAIN_ID_OPBNB = 204


class BnbChainProvider:
    """
    BnbChainProvider is a provider for the BNB Chain API.
    """

    address: ChecksumAddress
    current_client: Web3
    clients: dict[SupportedChain, Web3]

    def __init__(self, private_key: str, bsc_provider_url: str, opbnb_provider_url: Optional[str] = None):
        w3 = Web3(Web3.HTTPProvider(bsc_provider_url))
        if not w3.is_connected():
            raise ValueError(f'Failed to connect to BSC provider. BSC provider URL: {bsc_provider_url}')

        if w3.eth.chain_id != CHAIN_ID_BSC:
            raise ValueError(f'Provider is not connected to BSC. Chain ID: {w3.eth.chain_id}')

        account = w3.eth.account.from_key(private_key)
        self.address = account.address

        w3.middleware_onion.inject(ExtraDataToPOAMiddleware, layer=0)
        w3.middleware_onion.add(SignAndSendRawMiddlewareBuilder.build(account))  # type: ignore
        w3.eth.default_account = self.address

        self.current_client = w3
        self.clients = {
            SupportedChain.BSC: w3,
        }

        if opbnb_provider_url:
            w3 = Web3(Web3.HTTPProvider(opbnb_provider_url))
            if not w3.is_connected():
                raise ValueError(f'Failed to connect to opBNB provider. opBNB provider URL: {opbnb_provider_url}')

            if w3.eth.chain_id != CHAIN_ID_OPBNB:
                raise ValueError(f'Provider is not connected to opBNB. Chain ID: {w3.eth.chain_id}')

            w3.middleware_onion.inject(SignAndSendRawMiddlewareBuilder.build(account), layer=0)  # type: ignore
            w3.eth.default_account = account.address

            self.clients[SupportedChain.OPBNB] = w3

    def get_address(self) -> ChecksumAddress:
        return self.address

    def get_client(self, chain: SupportedChain) -> Web3:
        client = self.clients.get(chain)
        if client is None:
            raise ValueError(f'Client for chain {chain} not found')
        return client

    def get_current_client(self) -> Web3:
        return self.current_client

    def switch_client(self, chain: SupportedChain):
        if chain not in self.clients:
            raise ValueError(f'Client for chain {chain} not found')

        self.current_client = self.clients[chain]



from collections.abc import Callable, Coroutine
from typing import Any, Optional

from pydantic import BaseModel


class BnbChainAction(BaseModel):
    """BNB Chain Action Base Class."""

    name: str
    description: str
    args_schema: Optional[type[BaseModel]] = None
    func: Optional[Callable[..., str]] = None
    async_func: Optional[Callable[..., Coroutine[Any, Any, str]]] = None



import logging
from collections.abc import Callable
from typing import Optional

from pydantic import BaseModel, Field
from web3 import Web3

from bnb_chain_agentkit.actions.abi import ERC20_ABI
from bnb_chain_agentkit.actions.bnb_chain_action import BnbChainAction
from bnb_chain_agentkit.provider.bnb_chain_provider import BnbChainProvider

logger = logging.getLogger(__name__)

GET_BALANCE_PROMPT = """
This tool retrieves the balance of a specified asset for a given account.

It takes the account address and the token contract address as inputs.

Important notes:
- The account address is optional. If not specified, the default account will be used.
- The token could be a valid token address or "BNB".
"""


class GetBalanceInput(BaseModel):
    """Input argument schema for get balance action."""

    account: Optional[str] = Field(
        None,
        description='The account address to get the balance for. Optional. Could be None or a valid ethereum address starting with "0x".',
    )
    token: str = Field(
        'BNB',
        description='The token to get the balance for, e.g. `BNB` or `0x...`.',
    )


def get_balance(provider: BnbChainProvider, account: Optional[str], token: str) -> str:
    """Get balance of a specified asset for a given account.

    Args:
        provider (BnbChainProvider): The provider to use for the query.
        account (Optional[str]): The account address to get the balance for. Optional. Could be None or a valid ethereum address starting with "0x".',
        token (str): The token to get the balance for, e.g. `BNB` or `0x...`.

    Returns:
        str: A message containing the balance and decimals of the account for the given asset.
    """

    client = provider.get_current_client()
    address = Web3.to_checksum_address(account) if account else provider.get_address()

    if token == 'BNB':
        balance = client.eth.get_balance(address)  # type: ignore[arg-type]
        decimals = 18
        token_name = token
    else:
        token_address = Web3.to_checksum_address(token)
        contract = client.eth.contract(token_address, abi=ERC20_ABI)

        balance = contract.functions.balanceOf(address).call()
        decimals = contract.functions.decimals().call()
        token_name = contract.functions.name().call()

    return f'Balances for account {address} of {token_name}:\n{balance} (decimals: {decimals})'


class GetBalanceAction(BnbChainAction):
    """Get wallet balance action."""

    name: str = 'get_balance'
    description: str = GET_BALANCE_PROMPT
    args_schema: Optional[type[BaseModel]] = GetBalanceInput
    func: Optional[Callable[..., str]] = get_balance



ERC20_ABI = """
[
  {
    "type": "constructor",
    "inputs": [
      {
        "name": "name_",
        "type": "string",
        "internalType": "string"
      },
      {
        "name": "symbol_",
        "type": "string",
        "internalType": "string"
      }
    ],
    "stateMutability": "nonpayable"
  },
  {
    "type": "function",
    "name": "allowance",
    "inputs": [
      {
        "name": "owner",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "spender",
        "type": "address",
        "internalType": "address"
      }
    ],
    "outputs": [
      {
        "name": "",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "stateMutability": "view"
  },
  {
    "type": "function",
    "name": "approve",
    "inputs": [
      {
        "name": "spender",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "amount",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "outputs": [
      {
        "name": "",
        "type": "bool",
        "internalType": "bool"
      }
    ],
    "stateMutability": "nonpayable"
  },
  {
    "type": "function",
    "name": "balanceOf",
    "inputs": [
      {
        "name": "account",
        "type": "address",
        "internalType": "address"
      }
    ],
    "outputs": [
      {
        "name": "",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "stateMutability": "view"
  },
  {
    "type": "function",
    "name": "decimals",
    "inputs": [],
    "outputs": [
      {
        "name": "",
        "type": "uint8",
        "internalType": "uint8"
      }
    ],
    "stateMutability": "view"
  },
  {
    "type": "function",
    "name": "decreaseAllowance",
    "inputs": [
      {
        "name": "spender",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "subtractedValue",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "outputs": [
      {
        "name": "",
        "type": "bool",
        "internalType": "bool"
      }
    ],
    "stateMutability": "nonpayable"
  },
  {
    "type": "function",
    "name": "increaseAllowance",
    "inputs": [
      {
        "name": "spender",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "addedValue",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "outputs": [
      {
        "name": "",
        "type": "bool",
        "internalType": "bool"
      }
    ],
    "stateMutability": "nonpayable"
  },
  {
    "type": "function",
    "name": "name",
    "inputs": [],
    "outputs": [
      {
        "name": "",
        "type": "string",
        "internalType": "string"
      }
    ],
    "stateMutability": "view"
  },
  {
    "type": "function",
    "name": "symbol",
    "inputs": [],
    "outputs": [
      {
        "name": "",
        "type": "string",
        "internalType": "string"
      }
    ],
    "stateMutability": "view"
  },
  {
    "type": "function",
    "name": "totalSupply",
    "inputs": [],
    "outputs": [
      {
        "name": "",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "stateMutability": "view"
  },
  {
    "type": "function",
    "name": "transfer",
    "inputs": [
      {
        "name": "to",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "amount",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "outputs": [
      {
        "name": "",
        "type": "bool",
        "internalType": "bool"
      }
    ],
    "stateMutability": "nonpayable"
  },
  {
    "type": "function",
    "name": "transferFrom",
    "inputs": [
      {
        "name": "from",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "to",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "amount",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "outputs": [
      {
        "name": "",
        "type": "bool",
        "internalType": "bool"
      }
    ],
    "stateMutability": "nonpayable"
  },
  {
    "type": "event",
    "name": "Approval",
    "inputs": [
      {
        "name": "owner",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "spender",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "value",
        "type": "uint256",
        "indexed": false,
        "internalType": "uint256"
      }
    ],
    "anonymous": false
  },
  {
    "type": "event",
    "name": "Transfer",
    "inputs": [
      {
        "name": "from",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "to",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "value",
        "type": "uint256",
        "indexed": false,
        "internalType": "uint256"
      }
    ],
    "anonymous": false
  }
]
"""

L1_STANDARD_BRIDGE_ABI = """
[
  {
    "type": "constructor",
    "inputs": [],
    "stateMutability": "nonpayable"
  },
  {
    "type": "receive",
    "stateMutability": "payable"
  },
  {
    "type": "function",
    "name": "MESSENGER",
    "inputs": [],
    "outputs": [
      {
        "name": "",
        "type": "address",
        "internalType": "contract CrossDomainMessenger"
      }
    ],
    "stateMutability": "view"
  },
  {
    "type": "function",
    "name": "OTHER_BRIDGE",
    "inputs": [],
    "outputs": [
      {
        "name": "",
        "type": "address",
        "internalType": "contract StandardBridge"
      }
    ],
    "stateMutability": "view"
  },
  {
    "type": "function",
    "name": "bridgeERC20",
    "inputs": [
      {
        "name": "_localToken",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_remoteToken",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_amount",
        "type": "uint256",
        "internalType": "uint256"
      },
      {
        "name": "_minGasLimit",
        "type": "uint32",
        "internalType": "uint32"
      },
      {
        "name": "_extraData",
        "type": "bytes",
        "internalType": "bytes"
      }
    ],
    "outputs": [],
    "stateMutability": "nonpayable"
  },
  {
    "type": "function",
    "name": "bridgeERC20To",
    "inputs": [
      {
        "name": "_localToken",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_remoteToken",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_to",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_amount",
        "type": "uint256",
        "internalType": "uint256"
      },
      {
        "name": "_minGasLimit",
        "type": "uint32",
        "internalType": "uint32"
      },
      {
        "name": "_extraData",
        "type": "bytes",
        "internalType": "bytes"
      }
    ],
    "outputs": [],
    "stateMutability": "nonpayable"
  },
  {
    "type": "function",
    "name": "bridgeETH",
    "inputs": [
      {
        "name": "_minGasLimit",
        "type": "uint32",
        "internalType": "uint32"
      },
      {
        "name": "_extraData",
        "type": "bytes",
        "internalType": "bytes"
      }
    ],
    "outputs": [],
    "stateMutability": "payable"
  },
  {
    "type": "function",
    "name": "bridgeETHTo",
    "inputs": [
      {
        "name": "_to",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_minGasLimit",
        "type": "uint32",
        "internalType": "uint32"
      },
      {
        "name": "_extraData",
        "type": "bytes",
        "internalType": "bytes"
      }
    ],
    "outputs": [],
    "stateMutability": "payable"
  },
  {
    "type": "function",
    "name": "depositERC20",
    "inputs": [
      {
        "name": "_l1Token",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_l2Token",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_amount",
        "type": "uint256",
        "internalType": "uint256"
      },
      {
        "name": "_minGasLimit",
        "type": "uint32",
        "internalType": "uint32"
      },
      {
        "name": "_extraData",
        "type": "bytes",
        "internalType": "bytes"
      }
    ],
    "outputs": [],
    "stateMutability": "nonpayable"
  },
  {
    "type": "function",
    "name": "depositERC20To",
    "inputs": [
      {
        "name": "_l1Token",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_l2Token",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_to",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_amount",
        "type": "uint256",
        "internalType": "uint256"
      },
      {
        "name": "_minGasLimit",
        "type": "uint32",
        "internalType": "uint32"
      },
      {
        "name": "_extraData",
        "type": "bytes",
        "internalType": "bytes"
      }
    ],
    "outputs": [],
    "stateMutability": "nonpayable"
  },
  {
    "type": "function",
    "name": "depositETH",
    "inputs": [
      {
        "name": "_minGasLimit",
        "type": "uint32",
        "internalType": "uint32"
      },
      {
        "name": "_extraData",
        "type": "bytes",
        "internalType": "bytes"
      }
    ],
    "outputs": [],
    "stateMutability": "payable"
  },
  {
    "type": "function",
    "name": "depositETHTo",
    "inputs": [
      {
        "name": "_to",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_minGasLimit",
        "type": "uint32",
        "internalType": "uint32"
      },
      {
        "name": "_extraData",
        "type": "bytes",
        "internalType": "bytes"
      }
    ],
    "outputs": [],
    "stateMutability": "payable"
  },
  {
    "type": "function",
    "name": "deposits",
    "inputs": [
      {
        "name": "",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "",
        "type": "address",
        "internalType": "address"
      }
    ],
    "outputs": [
      {
        "name": "",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "stateMutability": "view"
  },
  {
    "type": "function",
    "name": "finalizeBridgeERC20",
    "inputs": [
      {
        "name": "_localToken",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_remoteToken",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_from",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_to",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_amount",
        "type": "uint256",
        "internalType": "uint256"
      },
      {
        "name": "_extraData",
        "type": "bytes",
        "internalType": "bytes"
      }
    ],
    "outputs": [],
    "stateMutability": "nonpayable"
  },
  {
    "type": "function",
    "name": "finalizeBridgeETH",
    "inputs": [
      {
        "name": "_from",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_to",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_amount",
        "type": "uint256",
        "internalType": "uint256"
      },
      {
        "name": "_extraData",
        "type": "bytes",
        "internalType": "bytes"
      }
    ],
    "outputs": [],
    "stateMutability": "payable"
  },
  {
    "type": "function",
    "name": "finalizeERC20Withdrawal",
    "inputs": [
      {
        "name": "_l1Token",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_l2Token",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_from",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_to",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_amount",
        "type": "uint256",
        "internalType": "uint256"
      },
      {
        "name": "_extraData",
        "type": "bytes",
        "internalType": "bytes"
      }
    ],
    "outputs": [],
    "stateMutability": "nonpayable"
  },
  {
    "type": "function",
    "name": "finalizeETHWithdrawal",
    "inputs": [
      {
        "name": "_from",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_to",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "_amount",
        "type": "uint256",
        "internalType": "uint256"
      },
      {
        "name": "_extraData",
        "type": "bytes",
        "internalType": "bytes"
      }
    ],
    "outputs": [],
    "stateMutability": "payable"
  },
  {
    "type": "function",
    "name": "initialize",
    "inputs": [
      {
        "name": "_messenger",
        "type": "address",
        "internalType": "contract CrossDomainMessenger"
      },
      {
        "name": "_superchainConfig",
        "type": "address",
        "internalType": "contract SuperchainConfig"
      },
      {
        "name": "_systemConfig",
        "type": "address",
        "internalType": "contract SystemConfig"
      }
    ],
    "outputs": [],
    "stateMutability": "nonpayable"
  },
  {
    "type": "function",
    "name": "l2TokenBridge",
    "inputs": [],
    "outputs": [
      {
        "name": "",
        "type": "address",
        "internalType": "address"
      }
    ],
    "stateMutability": "view"
  },
  {
    "type": "function",
    "name": "messenger",
    "inputs": [],
    "outputs": [
      {
        "name": "",
        "type": "address",
        "internalType": "contract CrossDomainMessenger"
      }
    ],
    "stateMutability": "view"
  },
  {
    "type": "function",
    "name": "otherBridge",
    "inputs": [],
    "outputs": [
      {
        "name": "",
        "type": "address",
        "internalType": "contract StandardBridge"
      }
    ],
    "stateMutability": "view"
  },
  {
    "type": "function",
    "name": "paused",
    "inputs": [],
    "outputs": [
      {
        "name": "",
        "type": "bool",
        "internalType": "bool"
      }
    ],
    "stateMutability": "view"
  },
  {
    "type": "function",
    "name": "superchainConfig",
    "inputs": [],
    "outputs": [
      {
        "name": "",
        "type": "address",
        "internalType": "contract SuperchainConfig"
      }
    ],
    "stateMutability": "view"
  },
  {
    "type": "function",
    "name": "systemConfig",
    "inputs": [],
    "outputs": [
      {
        "name": "",
        "type": "address",
        "internalType": "contract SystemConfig"
      }
    ],
    "stateMutability": "view"
  },
  {
    "type": "function",
    "name": "version",
    "inputs": [],
    "outputs": [
      {
        "name": "",
        "type": "string",
        "internalType": "string"
      }
    ],
    "stateMutability": "view"
  },
  {
    "type": "event",
    "name": "ERC20BridgeFinalized",
    "inputs": [
      {
        "name": "localToken",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "remoteToken",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "from",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "to",
        "type": "address",
        "indexed": false,
        "internalType": "address"
      },
      {
        "name": "amount",
        "type": "uint256",
        "indexed": false,
        "internalType": "uint256"
      },
      {
        "name": "extraData",
        "type": "bytes",
        "indexed": false,
        "internalType": "bytes"
      }
    ],
    "anonymous": false
  },
  {
    "type": "event",
    "name": "ERC20BridgeInitiated",
    "inputs": [
      {
        "name": "localToken",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "remoteToken",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "from",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "to",
        "type": "address",
        "indexed": false,
        "internalType": "address"
      },
      {
        "name": "amount",
        "type": "uint256",
        "indexed": false,
        "internalType": "uint256"
      },
      {
        "name": "extraData",
        "type": "bytes",
        "indexed": false,
        "internalType": "bytes"
      }
    ],
    "anonymous": false
  },
  {
    "type": "event",
    "name": "ERC20DepositInitiated",
    "inputs": [
      {
        "name": "l1Token",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "l2Token",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "from",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "to",
        "type": "address",
        "indexed": false,
        "internalType": "address"
      },
      {
        "name": "amount",
        "type": "uint256",
        "indexed": false,
        "internalType": "uint256"
      },
      {
        "name": "extraData",
        "type": "bytes",
        "indexed": false,
        "internalType": "bytes"
      }
    ],
    "anonymous": false
  },
  {
    "type": "event",
    "name": "ERC20WithdrawalFinalized",
    "inputs": [
      {
        "name": "l1Token",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "l2Token",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "from",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "to",
        "type": "address",
        "indexed": false,
        "internalType": "address"
      },
      {
        "name": "amount",
        "type": "uint256",
        "indexed": false,
        "internalType": "uint256"
      },
      {
        "name": "extraData",
        "type": "bytes",
        "indexed": false,
        "internalType": "bytes"
      }
    ],
    "anonymous": false
  },
  {
    "type": "event",
    "name": "ETHBridgeFinalized",
    "inputs": [
      {
        "name": "from",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "to",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "amount",
        "type": "uint256",
        "indexed": false,
        "internalType": "uint256"
      },
      {
        "name": "extraData",
        "type": "bytes",
        "indexed": false,
        "internalType": "bytes"
      }
    ],
    "anonymous": false
  },
  {
    "type": "event",
    "name": "ETHBridgeInitiated",
    "inputs": [
      {
        "name": "from",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "to",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "amount",
        "type": "uint256",
        "indexed": false,
        "internalType": "uint256"
      },
      {
        "name": "extraData",
        "type": "bytes",
        "indexed": false,
        "internalType": "bytes"
      }
    ],
    "anonymous": false
  },
  {
    "type": "event",
    "name": "ETHDepositInitiated",
    "inputs": [
      {
        "name": "from",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "to",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "amount",
        "type": "uint256",
        "indexed": false,
        "internalType": "uint256"
      },
      {
        "name": "extraData",
        "type": "bytes",
        "indexed": false,
        "internalType": "bytes"
      }
    ],
    "anonymous": false
  },
  {
    "type": "event",
    "name": "ETHWithdrawalFinalized",
    "inputs": [
      {
        "name": "from",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "to",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "amount",
        "type": "uint256",
        "indexed": false,
        "internalType": "uint256"
      },
      {
        "name": "extraData",
        "type": "bytes",
        "indexed": false,
        "internalType": "bytes"
      }
    ],
    "anonymous": false
  },
  {
    "type": "event",
    "name": "Initialized",
    "inputs": [
      {
        "name": "version",
        "type": "uint8",
        "indexed": false,
        "internalType": "uint8"
      }
    ],
    "anonymous": false
  }
]
"""

L2_STANDARD_BRIDGE_ABI = """
[
  {
    "type": "constructor",
    "inputs": [
      {
        "name": "_owner",
        "type": "address",
        "internalType": "address payable"
      },
      {
        "name": "_delegationFee",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "stateMutability": "nonpayable"
  },
  {
    "name": "AddressEmptyCode",
    "type": "error",
    "inputs": [
      { "name": "target", "type": "address", "internalType": "address" }
    ]
  },
  {
    "name": "AddressInsufficientBalance",
    "type": "error",
    "inputs": [
      { "name": "account", "type": "address", "internalType": "address" }
    ]
  },
  { "name": "FailedInnerCall", "type": "error", "inputs": [] },
  {
    "name": "OwnableInvalidOwner",
    "type": "error",
    "inputs": [
      { "name": "owner", "type": "address", "internalType": "address" }
    ]
  },
  {
    "name": "OwnableUnauthorizedAccount",
    "type": "error",
    "inputs": [
      { "name": "account", "type": "address", "internalType": "address" }
    ]
  },
  {
    "name": "SafeERC20FailedOperation",
    "type": "error",
    "inputs": [
      { "name": "token", "type": "address", "internalType": "address" }
    ]
  },
  {
    "name": "OwnershipTransferred",
    "type": "event",
    "inputs": [
      {
        "name": "previousOwner",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "newOwner",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      }
    ],
    "anonymous": false,
    "signature": "0x8be0079c531659141344cd1fd0a4f28419497f9722a3daafe3b4186f6b6457e0"
  },
  {
    "name": "SetDelegationFee",
    "type": "event",
    "inputs": [
      {
        "name": "_delegationFee",
        "type": "uint256",
        "indexed": false,
        "internalType": "uint256"
      }
    ],
    "anonymous": false,
    "signature": "0x0322f3257c2afe5fe8da7ab561f0d3384148487412fe2751678f2188731c0815"
  },
  {
    "name": "WithdrawTo",
    "type": "event",
    "inputs": [
      {
        "name": "from",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "l2Token",
        "type": "address",
        "indexed": true,
        "internalType": "address"
      },
      {
        "name": "to",
        "type": "address",
        "indexed": false,
        "internalType": "address"
      },
      {
        "name": "amount",
        "type": "uint256",
        "indexed": false,
        "internalType": "uint256"
      },
      {
        "name": "minGasLimit",
        "type": "uint32",
        "indexed": false,
        "internalType": "uint32"
      },
      {
        "name": "extraData",
        "type": "bytes",
        "indexed": false,
        "internalType": "bytes"
      }
    ],
    "anonymous": false,
    "signature": "0x56f66275d9ebc94b7d6895aa0d96a3783550d0183ba106408d387d19f2e877f1"
  },
  {
    "name": "L2_STANDARD_BRIDGE",
    "type": "function",
    "inputs": [],
    "outputs": [
      {
        "name": "",
        "type": "address",
        "value": "0x4200000000000000000000000000000000000010",
        "internalType": "contract IL2StandardBridge"
      }
    ],
    "constant": true,
    "signature": "0x21d12763",
    "stateMutability": "view"
  },
  {
    "name": "L2_STANDARD_BRIDGE_ADDRESS",
    "type": "function",
    "inputs": [],
    "outputs": [
      {
        "name": "",
        "type": "address",
        "value": "0x4200000000000000000000000000000000000010",
        "internalType": "address"
      }
    ],
    "constant": true,
    "signature": "0x2cb7cb06",
    "stateMutability": "view"
  },
  {
    "name": "delegationFee",
    "type": "function",
    "inputs": [],
    "outputs": [
      {
        "name": "",
        "type": "uint256",
        "value": "2000000000000000",
        "internalType": "uint256"
      }
    ],
    "constant": true,
    "signature": "0xc5f0a58f",
    "stateMutability": "view"
  },
  {
    "name": "owner",
    "type": "function",
    "inputs": [],
    "outputs": [
      {
        "name": "",
        "type": "address",
        "value": "0xCe4750fDc02A07Eb0d99cA798CD5c170D8F8410A",
        "internalType": "address"
      }
    ],
    "constant": true,
    "signature": "0x8da5cb5b",
    "stateMutability": "view"
  },
  {
    "name": "renounceOwnership",
    "type": "function",
    "inputs": [],
    "outputs": [],
    "signature": "0x715018a6",
    "stateMutability": "nonpayable"
  },
  {
    "name": "setDelegationFee",
    "type": "function",
    "inputs": [
      {
        "name": "_delegationFee",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "outputs": [],
    "signature": "0x55bfc81c",
    "stateMutability": "nonpayable"
  },
  {
    "name": "transferOwnership",
    "type": "function",
    "inputs": [
      { "name": "newOwner", "type": "address", "internalType": "address" }
    ],
    "outputs": [],
    "signature": "0xf2fde38b",
    "stateMutability": "nonpayable"
  },
  {
    "name": "withdraw",
    "type": "function",
    "inputs": [
      { "name": "_l2Token", "type": "address", "internalType": "address" },
      { "name": "_amount", "type": "uint256", "internalType": "uint256" },
      { "name": "_minGasLimit", "type": "uint32", "internalType": "uint32" },
      { "name": "_extraData", "type": "bytes", "internalType": "bytes" }
    ],
    "outputs": [],
    "payable": true,
    "signature": "0x32b7006d",
    "stateMutability": "payable"
  },
  {
    "name": "withdrawFee",
    "type": "function",
    "inputs": [
      { "name": "_recipient", "type": "address", "internalType": "address" }
    ],
    "outputs": [],
    "signature": "0x1ac3ddeb",
    "stateMutability": "nonpayable"
  },
  {
    "name": "withdrawFeeToL1",
    "type": "function",
    "inputs": [
      { "name": "_recipient", "type": "address", "internalType": "address" },
      { "name": "_minGasLimit", "type": "uint32", "internalType": "uint32" },
      { "name": "_extraData", "type": "bytes", "internalType": "bytes" }
    ],
    "outputs": [],
    "signature": "0x244cafe0",
    "stateMutability": "nonpayable"
  },
  {
    "name": "withdrawTo",
    "type": "function",
    "inputs": [
      { "name": "_l2Token", "type": "address", "internalType": "address" },
      { "name": "_to", "type": "address", "internalType": "address" },
      { "name": "_amount", "type": "uint256", "internalType": "uint256" },
      { "name": "_minGasLimit", "type": "uint32", "internalType": "uint32" },
      { "name": "_extraData", "type": "bytes", "internalType": "bytes" }
    ],
    "outputs": [],
    "payable": true,
    "signature": "0xa3a79548",
    "stateMutability": "payable"
  }
]
"""

LISTA_DAO_ABI = """
[
  { "inputs": [], "stateMutability": "nonpayable", "type": "constructor" },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_account",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "ClaimAllWithdrawals",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_uuid",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "ClaimUndelegated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_validator",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_uuid",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "ClaimUndelegatedFrom",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_account",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_idx",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "ClaimWithdrawal",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "Delegate",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "_validator",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "_delegateVotePower",
        "type": "bool"
      }
    ],
    "name": "DelegateTo",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "_delegateTo",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_votesChange",
        "type": "uint256"
      }
    ],
    "name": "DelegateVoteTo",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "_src",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "Deposit",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_address",
        "type": "address"
      }
    ],
    "name": "DisableValidator",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint8",
        "name": "version",
        "type": "uint8"
      }
    ],
    "name": "Initialized",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "Paused",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_address",
        "type": "address"
      }
    ],
    "name": "ProposeManager",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "_src",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "_dest",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "ReDelegate",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_rewardsId",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "Redelegate",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_address",
        "type": "address"
      }
    ],
    "name": "RemoveValidator",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_account",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amountInSlisBnb",
        "type": "uint256"
      }
    ],
    "name": "RequestWithdraw",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "RewardsCompounded",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      },
      {
        "indexed": true,
        "internalType": "bytes32",
        "name": "previousAdminRole",
        "type": "bytes32"
      },
      {
        "indexed": true,
        "internalType": "bytes32",
        "name": "newAdminRole",
        "type": "bytes32"
      }
    ],
    "name": "RoleAdminChanged",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "account",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "sender",
        "type": "address"
      }
    ],
    "name": "RoleGranted",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "account",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "sender",
        "type": "address"
      }
    ],
    "name": "RoleRevoked",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_annualRate",
        "type": "uint256"
      }
    ],
    "name": "SetAnnualRate",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_address",
        "type": "address"
      }
    ],
    "name": "SetBSCValidator",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_address",
        "type": "address"
      }
    ],
    "name": "SetManager",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_minBnb",
        "type": "uint256"
      }
    ],
    "name": "SetMinBnb",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_address",
        "type": "address"
      }
    ],
    "name": "SetRedirectAddress",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "SetReserveAmount",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_address",
        "type": "address"
      }
    ],
    "name": "SetRevenuePool",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_synFee",
        "type": "uint256"
      }
    ],
    "name": "SetSynFee",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_validator",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "_credit",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "toRemove",
        "type": "bool"
      }
    ],
    "name": "SyncCreditContract",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_nextUndelegatedRequestIndex",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_bnbAmount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_shares",
        "type": "uint256"
      }
    ],
    "name": "Undelegate",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_operator",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_bnbAmount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_shares",
        "type": "uint256"
      }
    ],
    "name": "UndelegateFrom",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "UndelegateReserve",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "Unpaused",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_address",
        "type": "address"
      }
    ],
    "name": "WhitelistValidator",
    "type": "event"
  },
  {
    "inputs": [],
    "name": "BOT",
    "outputs": [{ "internalType": "bytes32", "name": "", "type": "bytes32" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "DEFAULT_ADMIN_ROLE",
    "outputs": [{ "internalType": "bytes32", "name": "", "type": "bytes32" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "GUARDIAN",
    "outputs": [{ "internalType": "bytes32", "name": "", "type": "bytes32" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "TEN_DECIMALS",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "acceptNewManager",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "amountToDelegate",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "annualRate",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_bnbAmount", "type": "uint256" }
    ],
    "name": "binarySearchCoveredMaxIndex",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_validator", "type": "address" }
    ],
    "name": "claimUndelegated",
    "outputs": [
      { "internalType": "uint256", "name": "_uuid", "type": "uint256" },
      { "internalType": "uint256", "name": "_amount", "type": "uint256" }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_idx", "type": "uint256" }
    ],
    "name": "claimWithdraw",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_user", "type": "address" },
      { "internalType": "uint256", "name": "_idx", "type": "uint256" }
    ],
    "name": "claimWithdrawFor",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "compoundRewards",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_operator", "type": "address" },
      { "internalType": "uint256", "name": "_bnbAmount", "type": "uint256" }
    ],
    "name": "convertBnbToShares",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_amount", "type": "uint256" }
    ],
    "name": "convertBnbToSnBnb",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_operator", "type": "address" },
      { "internalType": "uint256", "name": "_shares", "type": "uint256" }
    ],
    "name": "convertSharesToBnb",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "_amountInSlisBnb",
        "type": "uint256"
      }
    ],
    "name": "convertSnBnbToBnb",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "name": "creditContracts",
    "outputs": [{ "internalType": "address", "name": "", "type": "address" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "address", "name": "", "type": "address" }],
    "name": "creditStates",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_validator", "type": "address" },
      { "internalType": "uint256", "name": "_amount", "type": "uint256" }
    ],
    "name": "delegateTo",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "delegateVotePower",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_delegateTo", "type": "address" }
    ],
    "name": "delegateVoteTo",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "deposit",
    "outputs": [],
    "stateMutability": "payable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "depositReserve",
    "outputs": [],
    "stateMutability": "payable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_address", "type": "address" }
    ],
    "name": "disableValidator",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getAmountToUndelegate",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "_amountToUndelegate",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_uuid", "type": "uint256" }
    ],
    "name": "getBotUndelegateRequest",
    "outputs": [
      {
        "components": [
          {
            "internalType": "uint256",
            "name": "startTime",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "endTime",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "amount",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "amountInSnBnb",
            "type": "uint256"
          }
        ],
        "internalType": "struct IStakeManager.BotUndelegateRequest",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_validator", "type": "address" }
    ],
    "name": "getClaimableAmount",
    "outputs": [
      { "internalType": "uint256", "name": "_amount", "type": "uint256" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getContracts",
    "outputs": [
      { "internalType": "address", "name": "_manager", "type": "address" },
      { "internalType": "address", "name": "_slisBnb", "type": "address" },
      { "internalType": "address", "name": "_bscValidator", "type": "address" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_validator", "type": "address" }
    ],
    "name": "getDelegated",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_amount", "type": "uint256" }
    ],
    "name": "getRedelegateFee",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "bytes32", "name": "role", "type": "bytes32" }
    ],
    "name": "getRoleAdmin",
    "outputs": [{ "internalType": "bytes32", "name": "", "type": "bytes32" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getSlisBnbWithdrawLimit",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "_slisBnbWithdrawLimit",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getTotalBnbInValidators",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getTotalPooledBnb",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_user", "type": "address" },
      { "internalType": "uint256", "name": "_idx", "type": "uint256" }
    ],
    "name": "getUserRequestStatus",
    "outputs": [
      { "internalType": "bool", "name": "_isClaimable", "type": "bool" },
      { "internalType": "uint256", "name": "_amount", "type": "uint256" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_address", "type": "address" }
    ],
    "name": "getUserWithdrawalRequests",
    "outputs": [
      {
        "components": [
          { "internalType": "uint256", "name": "uuid", "type": "uint256" },
          {
            "internalType": "uint256",
            "name": "amountInSnBnb",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "startTime",
            "type": "uint256"
          }
        ],
        "internalType": "struct IStakeManager.WithdrawalRequest[]",
        "name": "",
        "type": "tuple[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "bytes32", "name": "role", "type": "bytes32" },
      { "internalType": "address", "name": "account", "type": "address" }
    ],
    "name": "grantRole",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "bytes32", "name": "role", "type": "bytes32" },
      { "internalType": "address", "name": "account", "type": "address" }
    ],
    "name": "hasRole",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_slisBnb", "type": "address" },
      { "internalType": "address", "name": "_admin", "type": "address" },
      { "internalType": "address", "name": "_manager", "type": "address" },
      { "internalType": "address", "name": "_bot", "type": "address" },
      { "internalType": "uint256", "name": "_synFee", "type": "uint256" },
      { "internalType": "address", "name": "_revenuePool", "type": "address" },
      { "internalType": "address", "name": "_validator", "type": "address" }
    ],
    "name": "initialize",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "minBnb",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "nextConfirmedRequestUUID",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "pause",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "paused",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "placeholder",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_address", "type": "address" }
    ],
    "name": "proposeNewManager",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "srcValidator", "type": "address" },
      { "internalType": "address", "name": "dstValidator", "type": "address" },
      { "internalType": "uint256", "name": "_amount", "type": "uint256" }
    ],
    "name": "redelegate",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "redirectAddress",
    "outputs": [{ "internalType": "address", "name": "", "type": "address" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_address", "type": "address" }
    ],
    "name": "removeValidator",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "bytes32", "name": "role", "type": "bytes32" },
      { "internalType": "address", "name": "account", "type": "address" }
    ],
    "name": "renounceRole",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "name": "requestIndexMap",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "requestUUID",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "_amountInSlisBnb",
        "type": "uint256"
      }
    ],
    "name": "requestWithdraw",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "reserveAmount",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "revenuePool",
    "outputs": [{ "internalType": "address", "name": "", "type": "address" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_address", "type": "address" }
    ],
    "name": "revokeBotRole",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "bytes32", "name": "role", "type": "bytes32" },
      { "internalType": "address", "name": "account", "type": "address" }
    ],
    "name": "revokeRole",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_annualRate", "type": "uint256" }
    ],
    "name": "setAnnualRate",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_address", "type": "address" }
    ],
    "name": "setBSCValidator",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_address", "type": "address" }
    ],
    "name": "setBotRole",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_amount", "type": "uint256" }
    ],
    "name": "setMinBnb",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_address", "type": "address" }
    ],
    "name": "setRedirectAddress",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "amount", "type": "uint256" }
    ],
    "name": "setReserveAmount",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_address", "type": "address" }
    ],
    "name": "setRevenuePool",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_synFee", "type": "uint256" }
    ],
    "name": "setSynFee",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "bytes4", "name": "interfaceId", "type": "bytes4" }
    ],
    "name": "supportsInterface",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "synFee",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "togglePause",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "toggleVote",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "totalDelegated",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "totalReserveAmount",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "unbondingBnb",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "undelegate",
    "outputs": [
      { "internalType": "uint256", "name": "_uuid", "type": "uint256" },
      { "internalType": "uint256", "name": "_amount", "type": "uint256" }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_operator", "type": "address" },
      { "internalType": "uint256", "name": "_amount", "type": "uint256" }
    ],
    "name": "undelegateFrom",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "_actualBnbAmount",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "undelegatedQuota",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "address", "name": "", "type": "address" }],
    "name": "validators",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "_address", "type": "address" }
    ],
    "name": "whitelistValidator",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "amount", "type": "uint256" }
    ],
    "name": "withdrawReserve",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  { "stateMutability": "payable", "type": "receive" }
]
"""

IPANCAKE_V3_ROUTER_ABI = """
[
  {
    "type": "function",
    "name": "exactInput",
    "inputs": [
      {
        "name": "params",
        "type": "tuple",
        "internalType": "struct ISwapRouter.ExactInputParams",
        "components": [
          {
            "name": "path",
            "type": "bytes",
            "internalType": "bytes"
          },
          {
            "name": "recipient",
            "type": "address",
            "internalType": "address"
          },
          {
            "name": "deadline",
            "type": "uint256",
            "internalType": "uint256"
          },
          {
            "name": "amountIn",
            "type": "uint256",
            "internalType": "uint256"
          },
          {
            "name": "amountOutMinimum",
            "type": "uint256",
            "internalType": "uint256"
          }
        ]
      }
    ],
    "outputs": [
      {
        "name": "amountOut",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "stateMutability": "payable"
  },
  {
    "type": "function",
    "name": "exactInputSingle",
    "inputs": [
      {
        "name": "params",
        "type": "tuple",
        "internalType": "struct ISwapRouter.ExactInputSingleParams",
        "components": [
          {
            "name": "tokenIn",
            "type": "address",
            "internalType": "address"
          },
          {
            "name": "tokenOut",
            "type": "address",
            "internalType": "address"
          },
          {
            "name": "fee",
            "type": "uint24",
            "internalType": "uint24"
          },
          {
            "name": "recipient",
            "type": "address",
            "internalType": "address"
          },
          {
            "name": "deadline",
            "type": "uint256",
            "internalType": "uint256"
          },
          {
            "name": "amountIn",
            "type": "uint256",
            "internalType": "uint256"
          },
          {
            "name": "amountOutMinimum",
            "type": "uint256",
            "internalType": "uint256"
          },
          {
            "name": "sqrtPriceLimitX96",
            "type": "uint160",
            "internalType": "uint160"
          }
        ]
      }
    ],
    "outputs": [
      {
        "name": "amountOut",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "stateMutability": "payable"
  },
  {
    "type": "function",
    "name": "exactOutput",
    "inputs": [
      {
        "name": "params",
        "type": "tuple",
        "internalType": "struct ISwapRouter.ExactOutputParams",
        "components": [
          {
            "name": "path",
            "type": "bytes",
            "internalType": "bytes"
          },
          {
            "name": "recipient",
            "type": "address",
            "internalType": "address"
          },
          {
            "name": "deadline",
            "type": "uint256",
            "internalType": "uint256"
          },
          {
            "name": "amountOut",
            "type": "uint256",
            "internalType": "uint256"
          },
          {
            "name": "amountInMaximum",
            "type": "uint256",
            "internalType": "uint256"
          }
        ]
      }
    ],
    "outputs": [
      {
        "name": "amountIn",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "stateMutability": "payable"
  },
  {
    "type": "function",
    "name": "exactOutputSingle",
    "inputs": [
      {
        "name": "params",
        "type": "tuple",
        "internalType": "struct ISwapRouter.ExactOutputSingleParams",
        "components": [
          {
            "name": "tokenIn",
            "type": "address",
            "internalType": "address"
          },
          {
            "name": "tokenOut",
            "type": "address",
            "internalType": "address"
          },
          {
            "name": "fee",
            "type": "uint24",
            "internalType": "uint24"
          },
          {
            "name": "recipient",
            "type": "address",
            "internalType": "address"
          },
          {
            "name": "deadline",
            "type": "uint256",
            "internalType": "uint256"
          },
          {
            "name": "amountOut",
            "type": "uint256",
            "internalType": "uint256"
          },
          {
            "name": "amountInMaximum",
            "type": "uint256",
            "internalType": "uint256"
          },
          {
            "name": "sqrtPriceLimitX96",
            "type": "uint160",
            "internalType": "uint160"
          }
        ]
      }
    ],
    "outputs": [
      {
        "name": "amountIn",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "stateMutability": "payable"
  },
  {
    "type": "function",
    "name": "pancakeV3SwapCallback",
    "inputs": [
      {
        "name": "amount0Delta",
        "type": "int256",
        "internalType": "int256"
      },
      {
        "name": "amount1Delta",
        "type": "int256",
        "internalType": "int256"
      },
      {
        "name": "data",
        "type": "bytes",
        "internalType": "bytes"
      }
    ],
    "outputs": [],
    "stateMutability": "nonpayable"
  }
]
"""

IPANCAKE_V3_FACTORY_ABI = """
[
  {
    "type": "function",
    "name": "getPool",
    "inputs": [
      {
        "name": "tokenA",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "tokenB",
        "type": "address",
        "internalType": "address"
      },
      {
        "name": "fee",
        "type": "uint24",
        "internalType": "uint24"
      }
    ],
    "outputs": [
      {
        "name": "pool",
        "type": "address",
        "internalType": "address"
      }
    ],
    "stateMutability": "view"
  }
]
"""

IPANCAKE_V3_POOL_ABI = """
[
  {
    "type": "function",
    "name": "liquidity",
    "inputs": [],
    "outputs": [
      {
        "name": "",
        "type": "uint128",
        "internalType": "uint128"
      }
    ],
    "stateMutability": "view"
  }
]
"""

IPANCAKE_QUOTER_V2_ABI = """
[
  {
    "type": "function",
    "name": "quoteExactInput",
    "inputs": [
      {
        "name": "path",
        "type": "bytes",
        "internalType": "bytes"
      },
      {
        "name": "amountIn",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "outputs": [
      {
        "name": "amountOut",
        "type": "uint256",
        "internalType": "uint256"
      },
      {
        "name": "sqrtPriceX96AfterList",
        "type": "uint160[]",
        "internalType": "uint160[]"
      },
      {
        "name": "initializedTicksCrossedList",
        "type": "uint32[]",
        "internalType": "uint32[]"
      },
      {
        "name": "gasEstimate",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "stateMutability": "nonpayable"
  },
  {
    "type": "function",
    "name": "quoteExactInputSingle",
    "inputs": [
      {
        "name": "params",
        "type": "tuple",
        "internalType": "struct IQuoterV2.QuoteExactInputSingleParams",
        "components": [
          {
            "name": "tokenIn",
            "type": "address",
            "internalType": "address"
          },
          {
            "name": "tokenOut",
            "type": "address",
            "internalType": "address"
          },
          {
            "name": "amountIn",
            "type": "uint256",
            "internalType": "uint256"
          },
          {
            "name": "fee",
            "type": "uint24",
            "internalType": "uint24"
          },
          {
            "name": "sqrtPriceLimitX96",
            "type": "uint160",
            "internalType": "uint160"
          }
        ]
      }
    ],
    "outputs": [
      {
        "name": "amountOut",
        "type": "uint256",
        "internalType": "uint256"
      },
      {
        "name": "sqrtPriceX96After",
        "type": "uint160",
        "internalType": "uint160"
      },
      {
        "name": "initializedTicksCrossed",
        "type": "uint32",
        "internalType": "uint32"
      },
      {
        "name": "gasEstimate",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "stateMutability": "nonpayable"
  },
  {
    "type": "function",
    "name": "quoteExactOutput",
    "inputs": [
      {
        "name": "path",
        "type": "bytes",
        "internalType": "bytes"
      },
      {
        "name": "amountOut",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "outputs": [
      {
        "name": "amountIn",
        "type": "uint256",
        "internalType": "uint256"
      },
      {
        "name": "sqrtPriceX96AfterList",
        "type": "uint160[]",
        "internalType": "uint160[]"
      },
      {
        "name": "initializedTicksCrossedList",
        "type": "uint32[]",
        "internalType": "uint32[]"
      },
      {
        "name": "gasEstimate",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "stateMutability": "nonpayable"
  },
  {
    "type": "function",
    "name": "quoteExactOutputSingle",
    "inputs": [
      {
        "name": "params",
        "type": "tuple",
        "internalType": "struct IQuoterV2.QuoteExactOutputSingleParams",
        "components": [
          {
            "name": "tokenIn",
            "type": "address",
            "internalType": "address"
          },
          {
            "name": "tokenOut",
            "type": "address",
            "internalType": "address"
          },
          {
            "name": "amount",
            "type": "uint256",
            "internalType": "uint256"
          },
          {
            "name": "fee",
            "type": "uint24",
            "internalType": "uint24"
          },
          {
            "name": "sqrtPriceLimitX96",
            "type": "uint160",
            "internalType": "uint160"
          }
        ]
      }
    ],
    "outputs": [
      {
        "name": "amountIn",
        "type": "uint256",
        "internalType": "uint256"
      },
      {
        "name": "sqrtPriceX96After",
        "type": "uint160",
        "internalType": "uint160"
      },
      {
        "name": "initializedTicksCrossed",
        "type": "uint32",
        "internalType": "uint32"
      },
      {
        "name": "gasEstimate",
        "type": "uint256",
        "internalType": "uint256"
      }
    ],
    "stateMutability": "nonpayable"
  }
]
"""

WBNB_ABI = """
[
    {
        "constant": true,
        "inputs": [],
        "name": "name",
        "outputs": [{ "name": "", "type": "string" }],
        "payable": false,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": false,
        "inputs": [
            { "name": "guy", "type": "address" },
            { "name": "wad", "type": "uint256" }
        ],
        "name": "approve",
        "outputs": [{ "name": "", "type": "bool" }],
        "payable": false,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": true,
        "inputs": [],
        "name": "totalSupply",
        "outputs": [{ "name": "", "type": "uint256" }],
        "payable": false,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": false,
        "inputs": [
            { "name": "src", "type": "address" },
            { "name": "dst", "type": "address" },
            { "name": "wad", "type": "uint256" }
        ],
        "name": "transferFrom",
        "outputs": [{ "name": "", "type": "bool" }],
        "payable": false,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": false,
        "inputs": [{ "name": "wad", "type": "uint256" }],
        "name": "withdraw",
        "outputs": [],
        "payable": false,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": true,
        "inputs": [],
        "name": "decimals",
        "outputs": [{ "name": "", "type": "uint8" }],
        "payable": false,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": true,
        "inputs": [{ "name": "", "type": "address" }],
        "name": "balanceOf",
        "outputs": [{ "name": "", "type": "uint256" }],
        "payable": false,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": true,
        "inputs": [],
        "name": "symbol",
        "outputs": [{ "name": "", "type": "string" }],
        "payable": false,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": false,
        "inputs": [
            { "name": "dst", "type": "address" },
            { "name": "wad", "type": "uint256" }
        ],
        "name": "transfer",
        "outputs": [{ "name": "", "type": "bool" }],
        "payable": false,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": false,
        "inputs": [],
        "name": "deposit",
        "outputs": [],
        "payable": true,
        "stateMutability": "payable",
        "type": "function"
    },
    {
        "constant": true,
        "inputs": [
            { "name": "", "type": "address" },
            { "name": "", "type": "address" }
        ],
        "name": "allowance",
        "outputs": [{ "name": "", "type": "uint256" }],
        "payable": false,
        "stateMutability": "view",
        "type": "function"
    },
    { "payable": true, "stateMutability": "payable", "type": "fallback" },
    {
        "anonymous": false,
        "inputs": [
            { "indexed": true, "name": "src", "type": "address" },
            { "indexed": true, "name": "guy", "type": "address" },
            { "indexed": false, "name": "wad", "type": "uint256" }
        ],
        "name": "Approval",
        "type": "event"
    },
    {
        "anonymous": false,
        "inputs": [
            { "indexed": true, "name": "src", "type": "address" },
            { "indexed": true, "name": "dst", "type": "address" },
            { "indexed": false, "name": "wad", "type": "uint256" }
        ],
        "name": "Transfer",
        "type": "event"
    },
    {
        "anonymous": false,
        "inputs": [
            { "indexed": true, "name": "dst", "type": "address" },
            { "indexed": false, "name": "wad", "type": "uint256" }
        ],
        "name": "Deposit",
        "type": "event"
    },
    {
        "anonymous": false,
        "inputs": [
            { "indexed": true, "name": "src", "type": "address" },
            { "indexed": false, "name": "wad", "type": "uint256" }
        ],
        "name": "Withdrawal",
        "type": "event"
    }
]
"""



import logging
from collections.abc import Callable
from enum import Enum
from typing import Optional

from pydantic import BaseModel, Field
from web3 import Web3

from bnb_chain_agentkit.actions.abi import ERC20_ABI, LISTA_DAO_ABI
from bnb_chain_agentkit.actions.bnb_chain_action import BnbChainAction
from bnb_chain_agentkit.provider.bnb_chain_provider import BnbChainProvider

logger = logging.getLogger(__name__)

LISTA_DAO_ADDRESS = Web3.to_checksum_address('0x1adB950d8bB3dA4bE104211D5AB038628e477fE6')
SLIS_BNB_ADDRESS = Web3.to_checksum_address('0xB0b84D294e0C75A6abe60171b70edEb2EFd14A1B')

STAKE_PROMPT = """
This tool helps stake BNB to Lista DAO on BSC.

It takes the action type and the amount as inputs.

Important notes:
- The DEPOSIT action could have several similar expressions, e.g. "delegate" or "stake".
- Amount is required for the DEPOSIT action.
- The WITHDRAW action could have several similar expressions, e.g. "undelegate" or "unstake".
- Amount is optional for the WITHDRAW action. If not provided, the entire balance will be withdrawn.
- Staking can only be done on the BSC mainnet.
- Ensure there is sufficient balance for the DEPOSIT action itself AND the gas cost of this action.
"""


class Action(Enum):
    DEPOSIT = 'deposit'
    WITHDRAW = 'withdraw'
    CLAIM = 'claim'


class StakeInput(BaseModel):
    """Input argument schema for stake action."""

    action: Action = Field(description='The action to perform.')
    amount: str = Field('0', description='The amount to perform the action on, e.g. `15`, `0.000001`.')


def stake(
    provider: BnbChainProvider,
    action: Action,
    amount: str,
) -> str:
    """Stake related actions on Lista DAO on BSC.

    Args:
        provider (BnbChainProvider): The provider to use for the transfer.
        action (Action): The action to perform.
        amount (str): The amount to perform the action on, e.g. `15`, `0.000001`.

    Returns:
        str: A message containing the action details.
    """

    client = provider.get_current_client()
    address = provider.get_address()

    if client.eth.chain_id != 56:
        return 'This tool can only be used on the BSC mainnet.'

    if action == Action.DEPOSIT:
        if amount == '0':
            return 'Amount is required for the "deposit" action.'
        return do_deposit(client, address, amount)
    elif action == Action.WITHDRAW:
        return do_withdraw(client, address, amount)
    elif action == Action.CLAIM:
        return do_claim(client, address)


def do_deposit(client: Web3, address: str, amount: str) -> str:
    amount_wei = Web3.to_wei(amount, 'ether')
    lista_dao = client.eth.contract(LISTA_DAO_ADDRESS, abi=LISTA_DAO_ABI)
    tx_hash = lista_dao.functions.deposit().transact({'value': amount_wei})
    client.eth.wait_for_transaction_receipt(tx_hash)

    slis_bnb = client.eth.contract(SLIS_BNB_ADDRESS, abi=ERC20_ABI)
    balance = slis_bnb.functions.balanceOf(address).call()
    balance_formatted = Web3.from_wei(balance, 'ether')

    return f'Deposited {amount} BNB to Lista DAO. Transaction hash: {tx_hash.to_0x_hex()}\nBalance of slisBNB: {balance_formatted}'


def do_withdraw(client: Web3, address: str, amount: str) -> str:
    slis_bnb = client.eth.contract(SLIS_BNB_ADDRESS, abi=ERC20_ABI)

    # If amount is None, withdraw the entire balance
    if amount == '0':
        amount_wei = slis_bnb.functions.balanceOf(address).call()
        amount = Web3.from_wei(amount_wei, 'ether')  # type: ignore
    else:
        amount_wei = Web3.to_wei(amount, 'ether')

    # check allowance
    allowance = slis_bnb.functions.allowance(address, LISTA_DAO_ADDRESS).call()
    if allowance < amount_wei:
        logger.info(f'Increasing allowance for Lista DAO. {amount_wei - allowance} more needed.')
        tx_hash = slis_bnb.functions.approve(LISTA_DAO_ADDRESS, amount_wei).transact()
        client.eth.wait_for_transaction_receipt(tx_hash)

    lista_dao = client.eth.contract(LISTA_DAO_ADDRESS, abi=LISTA_DAO_ABI)
    tx_hash = lista_dao.functions.requestWithdraw(amount_wei).transact()
    client.eth.wait_for_transaction_receipt(tx_hash)

    balance = slis_bnb.functions.balanceOf(address).call()
    balance_formatted = Web3.from_wei(balance, 'ether')

    return f'Requested to withdraw {amount} slisBNB from Lista DAO (Need 7 days to be claimed). Transaction hash: {tx_hash.to_0x_hex()}\nBalance of slisBNB: {balance_formatted}'


def do_claim(client: Web3, address: str) -> str:
    lista_dao = client.eth.contract(LISTA_DAO_ADDRESS, abi=LISTA_DAO_ABI)
    requests = lista_dao.functions.getUserWithdrawalRequests(address).call()

    txs = []
    total_claimed = 0
    for i in range(len(requests)):
        is_claimable, amount = lista_dao.functions.getUserRequestStatus(address, i).call()
        if is_claimable:
            tx_hash = lista_dao.functions.claimWithdraw(i).transact()
            client.eth.wait_for_transaction_receipt(tx_hash)
            txs.append(tx_hash)
            total_claimed += amount
        else:
            break

    total_claimed_formatted = Web3.from_wei(total_claimed, 'ether')

    return f'Claimed {total_claimed_formatted} slisBNB from Lista DAO. Transaction hashes: {", ".join([tx.to_0x_hex() for tx in txs])}'


class StakeAction(BnbChainAction):
    """Stake action."""

    name: str = 'stake'
    description: str = STAKE_PROMPT
    args_schema: Optional[type[BaseModel]] = StakeInput
    func: Optional[Callable[..., str]] = stake



import asyncio
import json
import logging
from collections.abc import Callable, Coroutine
from typing import Any, Optional

import websockets
from pydantic import BaseModel, Field
from websockets.exceptions import ConnectionClosed, WebSocketException

from bnb_chain_agentkit.actions.bnb_chain_action import BnbChainAction
from bnb_chain_agentkit.provider.bnb_chain_provider import BnbChainProvider

logger = logging.getLogger(__name__)

FAUCET_URL = 'wss://testnet.bnbchain.org/faucet-smart/api'
WEBSOCKET_TIMEOUT = 20  # seconds
MAX_RETRIES = 3

FAUCET_PROMPT = """
This tool will request test tokens from the BNB Chain official faucet.

It takes the token and recipient as inputs.

Important notes:
- The recipient is optional. If not specified, the default account will be used.
- The token is optional. If not specified, BNB will be used.
"""


class FaucetInput(BaseModel):
    """Input argument schema for faucet action."""

    token: str = Field(
        'BNB',
        description='The token to request, e.g. `BNB`, `BTC`, `BUSD`, `DAI`, `ETH`, `USDC`.',
    )
    recipient: Optional[str] = Field(
        None,
        description='The recipient to request the funds. Optional. Could be None or a valid ethereum address starting with "0x".',
    )


async def faucet(
    provider: BnbChainProvider,
    token: str,
    recipient: Optional[str],
) -> str:
    """Request test tokens from the BNB Chain official faucet.

    Args:
        provider (BnbChainProvider): The provider to use for the transfer.
        token (str): The token to request. Could be one of ["BNB", "BTC", "BUSD", "DAI", "ETH", "USDC"].
        recipient (Optional[str]): The recipient to request the funds. Optional. Could be None or a valid ethereum address starting with "0x".

    Returns:
        str: A message containing the transfer details.
    """

    recipient = recipient or provider.get_address()

    headers = {
        'Connection': 'Upgrade',
        'Upgrade': 'websocket',
    }
    message = {
        'tier': 0,
        'url': recipient,
        'symbol': token,
        'captcha': 'noCaptchaToken',
    }

    for attempt in range(MAX_RETRIES):
        try:
            async with websockets.connect(
                FAUCET_URL,
                extra_headers=headers,
                close_timeout=WEBSOCKET_TIMEOUT,
                ping_timeout=WEBSOCKET_TIMEOUT,
            ) as websocket:
                await websocket.send(json.dumps(message))

                try:
                    async with asyncio.timeout(WEBSOCKET_TIMEOUT):
                        while True:
                            response = await websocket.recv()
                            response_data = json.loads(response)

                            if 'error' in response_data:
                                return f'Error from faucet: {response_data["error"]}'

                            requests = response_data.get('requests', [])
                            if len(requests) > 0:
                                tx_hash = requests[0]['tx']['hash']
                                return f'Successfully requested {token} for {recipient}. Transaction hash: {tx_hash}'

                except asyncio.TimeoutError:
                    if attempt < MAX_RETRIES - 1:
                        continue
                    return 'Request timed out while waiting for faucet response'

        except ConnectionClosed:
            if attempt < MAX_RETRIES - 1:
                continue
            return 'Connection closed by faucet server'
        except WebSocketException as e:
            return f'WebSocket error: {str(e)}'
        except Exception as e:
            return f'Unexpected error: {str(e)}'

    return 'Failed to get response from faucet after maximum retries'


class FaucetAction(BnbChainAction):
    """Faucet action."""

    name: str = 'faucet'
    description: str = FAUCET_PROMPT
    args_schema: Optional[type[BaseModel]] = FaucetInput
    async_func: Optional[Callable[..., Coroutine[Any, Any, str]]] = faucet



import logging
import os
from collections.abc import Callable
from enum import Enum
from typing import Optional

from eth_typing import ChecksumAddress
from pydantic import BaseModel, Field
from solcx import compile_source, install_solc_pragma, set_solc_version_pragma
from solcx.exceptions import SolcNotInstalled
from web3 import Web3

from bnb_chain_agentkit.actions.utils.units import parse_units
from bnb_chain_agentkit.actions.bnb_chain_action import BnbChainAction
from bnb_chain_agentkit.provider.bnb_chain_provider import BnbChainProvider, SupportedChain

logger = logging.getLogger(__name__)

_solc_pragma = 'pragma solidity ^0.8.22;'

DEPLOY_PROMPT = """
This tool helps deploy an ERC contract on BSC or opBNB.

It takes contract type, token name, token symbol, initial supply, base uri and chain as inputs.

Important notes:
- For ERC20, token name and token symbol are required.
- For ERC721, token name, token symbol and base uri are required.
- For ERC1155, uri is required.
- The initial supply is in ether. If not specified, it is set to default value of 1000000000000000000.
"""


class ContractType(Enum):
    ERC20 = 'ERC20'
    ERC721 = 'ERC721'
    ERC1155 = 'ERC1155'


class DeployInput(BaseModel):
    """Input argument schema for deploy action."""

    contract_type: ContractType = Field(description='The contract type to deploy.')
    token_name: Optional[str] = Field(None, description='The token name. Must be provided for ERC20 and ERC721.')
    token_symbol: Optional[str] = Field(None, description='The token symbol. Must be provided for ERC20 and ERC721.')
    initial_supply: Optional[str] = Field('1000000000000000000', description='The initial supply.')
    base_uri: Optional[str] = Field(None, description='The base uri. Must be provided for ERC721 and ERC1155.')
    chain: SupportedChain = Field(SupportedChain.BSC, description='The chain to perform the deploy on.')


def deploy(
    provider: BnbChainProvider,
    contract_type: ContractType,
    token_name: Optional[str],
    token_symbol: Optional[str],
    initial_supply: Optional[str],
    base_uri: Optional[str],
    chain: SupportedChain,

) -> str:
    """Deploy an ERC contract on BSC or opBNB.

    Args:
        provider (BnbChainProvider): The provider to use for the deploy.
        contract_type (ContractType): The contract type to deploy.
        token_name (str): The token name.
        token_symbol (Optional[str]): The token symbol.
        initial_supply (Optional[str]): The initial supply.
        base_uri (Optional[str]): The base uri.
        chain (Optional[str]): The chain to perform the swap on.

    Returns:
        str: A message containing the action details.
    """

    client = provider.get_client(chain)
    account = provider.get_address()

    prepare_solc()
    if contract_type == ContractType.ERC20:
        if token_name is None:
            return 'Token name is required for ERC20.'
        if token_symbol is None:
            return 'Token symbol is required for ERC20.'
        
        initial_supply_wei = parse_units(initial_supply, 18) if initial_supply else 0
        print("initial_supply in ether is " + str(initial_supply) , "initial_supply_wei is " +str(initial_supply_wei))
        contract_address = deploy_contract(client, 'ERC20', token_name, token_symbol, initial_supply_wei, account)
    elif contract_type == ContractType.ERC721:
        if token_name is None:
            return 'Token name is required for ERC721.'
        if token_symbol is None:
            return 'Token symbol is required for ERC721.'
        if base_uri is None:
            return 'Base uri is required for ERC721.'

        contract_address = deploy_contract(client, 'ERC721', token_name, token_symbol, base_uri, account)
    elif contract_type == ContractType.ERC1155:
        if base_uri is None:
            return 'Base uri is required for ERC1155.'

        contract_address = deploy_contract(client, 'ERC1155', base_uri, account)

    return f'{contract_type.value} deployed successfully. Contract address: {contract_address}'


def deploy_contract(client: Web3, contract_name: str, *args) -> ChecksumAddress:
    current_dir = os.path.dirname(os.path.abspath(__file__))
    contract_path = os.path.join(current_dir, 'contracts', f'{contract_name}.sol')
    with open(contract_path, 'r') as c:
        compiled_sol = compile_source(
            c.read(),
            output_values=['abi', 'bin'],
            optimize=True,
            optimize_runs=200,
        )

    contract_interface = compiled_sol[f'<stdin>:{contract_name}Contract']
    bytecode = contract_interface['bin']
    abi = contract_interface['abi']

    if len(bytecode) == 0:
        raise ValueError('Bytecode is empty')

    contract = client.eth.contract(abi=abi, bytecode=bytecode)
    tx_hash = contract.constructor(*args).transact()
    receipt = client.eth.wait_for_transaction_receipt(tx_hash)
    return receipt.contractAddress  # type: ignore


def prepare_solc():
    try:
        set_solc_version_pragma(_solc_pragma)
    except SolcNotInstalled:
        logger.info('No compatible solc version installed. Installing...')
        install_solc_pragma(_solc_pragma, show_progress=True)


class DeployAction(BnbChainAction):
    """Deploy action."""

    name: str = 'deploy'
    description: str = DEPLOY_PROMPT
    args_schema: Optional[type[BaseModel]] = DeployInput
    func: Optional[Callable[..., str]] = deploy



"""Util that calls BNB Chain."""

import inspect
from collections.abc import Callable, Coroutine
from typing import Any

from langchain_core.utils import get_from_dict_or_env
from pydantic import BaseModel, model_validator

from bnb_chain_agentkit.provider.bnb_chain_provider import BnbChainProvider


class BnbChainAPIWrapper(BaseModel):
    """Wrapper for BNB Chain API."""

    provider: Any = None  #: :meta private:

    @model_validator(mode='before')
    @classmethod
    def validate_environment(cls, values: dict) -> Any:
        """Validate that the required env variables and python package exists in the environment and configure the Web3 provider."""
        private_key = get_from_dict_or_env(values, 'private_key', 'PRIVATE_KEY')
        bsc_provider_url = get_from_dict_or_env(values, 'bsc_provider_url', 'BSC_PROVIDER_URL')
        opbnb_provider_url = get_from_dict_or_env(values, 'opbnb_provider_url', 'OPBNB_PROVIDER_URL')

        provider = BnbChainProvider(private_key, bsc_provider_url, opbnb_provider_url)

        values['provider'] = provider

        return values

    def run_action(self, func: Callable[..., str], **kwargs) -> str:
        """Run a BNB Chain Action."""
        func_signature = inspect.signature(func)

        first_kwarg = next(iter(func_signature.parameters.values()), None)

        if first_kwarg and first_kwarg.annotation is BnbChainProvider:
            return func(self.provider, **kwargs)
        else:
            return func(**kwargs)

    async def async_run_action(self, async_func: Callable[..., Coroutine[Any, Any, str]], **kwargs) -> str:
        """Run a BNB Chain Action."""
        func_signature = inspect.signature(async_func)

        first_kwarg = next(iter(func_signature.parameters.values()), None)

        if first_kwarg and first_kwarg.annotation is BnbChainProvider:
            return await async_func(self.provider, **kwargs)
        else:
            return await async_func(**kwargs)



import logging
from collections.abc import Callable
from typing import Optional

from pydantic import BaseModel, Field
from web3 import Web3

from bnb_chain_agentkit.actions.abi import ERC20_ABI, L1_STANDARD_BRIDGE_ABI, L2_STANDARD_BRIDGE_ABI
from bnb_chain_agentkit.actions.bnb_chain_action import BnbChainAction
from bnb_chain_agentkit.actions.utils.units import parse_units
from bnb_chain_agentkit.provider.bnb_chain_provider import BnbChainProvider, SupportedChain

logger = logging.getLogger(__name__)

L1_BRIDGE_ADDRESS = Web3.to_checksum_address('0xF05F0e4362859c3331Cb9395CBC201E3Fa6757Ea')
L2_BRIDGE_ADDRESS = Web3.to_checksum_address('0x4000698e3De52120DE28181BaACda82B21568416')
LEGACY_ERC20_ETH_ADDRESS = Web3.to_checksum_address('0xDeadDeAddeAddEAddeadDEaDDEAdDeaDDeAD0000')

BRIDGE_PROMPT = """
This tool helps bridge tokens between BSC and opBNB.

It takes from_chain, to_chain, from_token, to_token, amount, and recipient as inputs.

Important notes:
- There are some similar expressions for the bridge action. For example:
    - "Bridge 1 BNB from BSC to opBNB"
    - "Deposit 1 BNB from BSC to opBNB"
    - "Withdraw 1 BNB from opBNB to BSC"
- from_chain and to_chain must be different.
- If from_token is not provided, the native asset (BNB) will be bridged. In this case, to_token must be None.
- When bridging from opBNB to BSC, the to_token should always be None.
- When bridging from opBNB to BSC, there is a delegation fee for the claim bot. Make sure to have enough funds to cover this fee.
"""


class BridgeInput(BaseModel):
    """Input argument schema for bridge action."""

    from_chain: SupportedChain = Field(description='The chain to bridge from.')
    to_chain: SupportedChain = Field(description='The chain to bridge to.')
    from_token: Optional[str] = Field(
        None,
        description='The token to bridge from. Optional. Could be None or a valid ethereum address starting with "0x".',
    )
    to_token: Optional[str] = Field(
        None,
        description='The token to bridge to. Optional. Could be None or a valid ethereum address starting with "0x".',
    )
    amount: str = Field(description='The amount to perform the action on, e.g. `15`, `0.000001`.')
    recipient: Optional[str] = Field(
        None,
        description='The recipient to bridge the funds to. Must be a valid Ethereum address starting with "0x". Optional. Could be None or a valid ethereum address starting with "0x".',
    )


def bridge(
    provider: BnbChainProvider,
    from_chain: SupportedChain,
    to_chain: SupportedChain,
    from_token: Optional[str],
    to_token: Optional[str],
    amount: str,
    recipient: Optional[str],
) -> str:
    """Bridge tokens between BSC and opBNB.

    Args:
        provider (BnbChainProvider): The provider to use for the bridge.
        from_chain (SupportedChain): The chain to bridge from.
        to_chain (SupportedChain): The chain to bridge to.
        from_token (Optional[str]): The token contract address to bridge from. Optional. Could be None or a valid ethereum address starting with "0x".
        to_token (Optional[str]): The token contract address to bridge to. Optional. Could be None or a valid ethereum address starting with "0x".
        amount (str): The amount to perform the action on, e.g. `15`, `0.000001`.
        recipient (Optional[str]): The recipient to bridge the funds to. Optional. Could be None or a valid ethereum address starting with "0x".

    Returns:
        str: A message containing the action details.
    """

    address = provider.get_address()
    recipient = Web3.to_checksum_address(recipient) if recipient else address

    from_token = from_token or 'BNB'
    self_bridge = recipient == address
    native_token_bridge = from_token == 'BNB'

    client = provider.get_client(from_chain)
    if from_chain == SupportedChain.BSC and to_chain == SupportedChain.OPBNB:
        l1_bridge = client.eth.contract(L1_BRIDGE_ADDRESS, abi=L1_STANDARD_BRIDGE_ABI)

        if native_token_bridge:
            amount_wei = Web3.to_wei(amount, 'ether')

            if self_bridge:
                tx_hash = l1_bridge.functions.depositETH(1, '0x').transact({'value': amount_wei})
            else:
                tx_hash = l1_bridge.functions.depositETHTo(recipient, 1, '0x').transact({'value': amount_wei})
        else:
            if to_token is None:
                raise ValueError('to_token must be provided when bridging ERC20 from BSC to opBNB')

            from_token = Web3.to_checksum_address(from_token)
            to_token = Web3.to_checksum_address(to_token)

            contract = client.eth.contract(from_token, abi=ERC20_ABI)
            decimals = contract.functions.decimals().call()
            amount_wei = parse_units(amount, decimals)

            # check ERC20 allowance
            allowance = contract.functions.allowance(address, L1_BRIDGE_ADDRESS).call()
            if allowance < amount_wei:
                logger.info(f'Increasing allowance for L1 bridge. {amount_wei - allowance} more needed.')
                tx_hash = contract.functions.approve(L1_BRIDGE_ADDRESS, amount_wei).transact()
                client.eth.wait_for_transaction_receipt(tx_hash)

            if self_bridge:
                tx_hash = l1_bridge.functions.depositERC20(from_token, to_token, amount_wei, 1, '0x').transact()
            else:
                tx_hash = l1_bridge.functions.depositERC20To(
                    from_token, to_token, recipient, amount_wei, 1, '0x'
                ).transact()

    elif from_chain == SupportedChain.OPBNB and to_chain == SupportedChain.BSC:
        l2_bridge = client.eth.contract(L2_BRIDGE_ADDRESS, abi=L2_STANDARD_BRIDGE_ABI)
        delegation_fee = l2_bridge.functions.delegationFee().call()

        if native_token_bridge:
            amount_wei = Web3.to_wei(amount, 'ether')
            value = amount_wei + delegation_fee

            if self_bridge:
                tx_hash = l2_bridge.functions.withdraw(LEGACY_ERC20_ETH_ADDRESS, amount_wei, 1, '0x').transact(
                    {'value': value}
                )
            else:
                tx_hash = l2_bridge.functions.withdrawTo(
                    LEGACY_ERC20_ETH_ADDRESS, recipient, amount_wei, 1, '0x'
                ).transact({'value': value})
        else:
            from_token = Web3.to_checksum_address(from_token)
            contract = client.eth.contract(from_token, abi=ERC20_ABI)
            decimals = contract.functions.decimals().call()
            amount_wei = parse_units(amount, decimals)

            # check ERC20 allowance
            allowance = contract.functions.allowance(address, L2_BRIDGE_ADDRESS).call()
            if allowance < amount_wei:
                logger.info(f'Increasing allowance for L2 bridge. {amount_wei - allowance} more needed.')
                tx_hash = contract.functions.approve(L2_BRIDGE_ADDRESS, amount_wei).transact()
                client.eth.wait_for_transaction_receipt(tx_hash)

            if self_bridge:
                tx_hash = l2_bridge.functions.withdraw(from_token, amount_wei, 1, '0x').transact(
                    {'value': delegation_fee}
                )
            else:
                tx_hash = l2_bridge.functions.withdrawTo(from_token, recipient, amount_wei, 1, '0x').transact(
                    {'value': delegation_fee}
                )

    else:
        raise ValueError(f'Invalid bridge direction: {from_chain} to {to_chain}')

    return f'Bridge {amount} {from_token} from {from_chain.value} to {to_chain.value}. Transaction hash: {tx_hash.to_0x_hex()}'


class BridgeAction(BnbChainAction):
    """Bridge action."""

    name: str = 'bridge'
    description: str = BRIDGE_PROMPT
    args_schema: Optional[type[BaseModel]] = BridgeInput
    func: Optional[Callable[..., str]] = bridge



from bnb_chain_agentkit.actions.bnb_chain_action import BnbChainAction
from bnb_chain_agentkit.actions.bridge import BridgeAction
from bnb_chain_agentkit.actions.deploy import DeployAction
from bnb_chain_agentkit.actions.faucet import FaucetAction
from bnb_chain_agentkit.actions.get_balance import GetBalanceAction
from bnb_chain_agentkit.actions.stake import StakeAction
from bnb_chain_agentkit.actions.swap import SwapAction
from bnb_chain_agentkit.actions.transfer import TransferAction

from bnb_chain_agentkit.actions.transaction_analytics import (
    GasFeeEstimationAction,
    TransactionReceiptAction,
    GasFeeEstimationAction,
    RecentTransactionsAction,
    PendingTransactionTrackerAction,
    GasPriceTrendAnalyzerAction,
    TransactionFeeEstimatorAction,
    TransactionVolumeAnalyzerAction,
    TokenTransferHistoryAction,
)
from bnb_chain_agentkit.actions.token_price import TokenPriceQueryAction


def get_all_bnb_chain_actions() -> list[type[BnbChainAction]]:
    """Retrieve all subclasses of BnbChainAction defined in the package."""
    actions = []
    for action in BnbChainAction.__subclasses__():
        actions.append(action())  # type: ignore
    return actions


BNB_CHAIN_ACTIONS = get_all_bnb_chain_actions()

__all__ = [
    'BNB_CHAIN_ACTIONS',
    'BnbChainAction',
    'GetBalanceAction',
    'TransferAction',
    'StakeAction',
    'FaucetAction',
    'BridgeAction',
    'DeployAction',
    'SwapAction',
    'GasFeeEstimationAction',
    'TransactionReceiptAction',
    'TokenPriceQueryAction',
    'RecentTransactionsAction',
    'PendingTransactionTrackerAction',
    'GasPriceTrendAnalyzerAction',
    'TransactionFeeEstimatorAction',
    'TransactionVolumeAnalyzerAction',
    'TokenTransferHistoryAction'
]



import os
import time
import requests
from pydantic import BaseModel, Field
from bnb_chain_agentkit.actions.bnb_chain_action import BnbChainAction
from bnb_chain_agentkit.provider.bnb_chain_provider import BnbChainProvider
from web3 import Web3

# A mapping for token symbols to contract addresses:
TOKEN_CONTRACT_ADDRESSES = {
    "BUSD": "0xe9e7cea3dedca5984780bafc599bd69add087d56",
    "BSC-USD":"0x55d398326f99059fF775485246999027B3197955", 
    "BUSD-T":"0x55d398326f99059ff775485246999027b3197955",
    "BEP20Ethereum":"0x2170Ed0880ac9A755fd29B2688956BD959F933F8",
    "ETH":"0x2170Ed0880ac9A755fd29B2688956BD959F933F8",
    "WBNB":"0xbb4CdB9CBd36B01bD1cBaEBF2De08d9173bc095c",
    "Wrapped BNB":"0xbb4CdB9CBd36B01bD1cBaEBF2De08d9173bc095c",
    "BTCB":"0x7130d2A12B9BCbFAe4f2634d864A1Ee1Ce3Ead9c",
    "BTCB Token":"0x7130d2A12B9BCbFAe4f2634d864A1Ee1Ce3Ead9c",
}

# --------------------------------------------------------------------
# 1. Gas Fee Estimation Tool
# --------------------------------------------------------------------

class GasFeeEstimationInput(BaseModel):
    # No input parameters needed.
    pass

def get_gas_fee(provider: BnbChainProvider) -> str:
    client = provider.get_current_client()
    gas_price = client.eth.gas_price  # in wei
    gas_price_gwei = Web3.from_wei(gas_price, 'gwei')
    return f"Current gas price is {gas_price_gwei} Gwei."

class GasFeeEstimationAction(BnbChainAction):
    name: str = "get_gas_fee"
    description: str = "Estimate current gas fee (in Gwei)."
    args_schema: type[BaseModel] = GasFeeEstimationInput
    func = get_gas_fee

# --------------------------------------------------------------------
# 2. Transaction Receipt / Analytics Tool
# --------------------------------------------------------------------

class TransactionReceiptInput(BaseModel):
    tx_hash: str = Field(..., description="Transaction hash to retrieve details for.")

def get_transaction_receipt(provider: BnbChainProvider, tx_hash: str) -> str:
    client = provider.get_current_client()
    try:
        receipt = client.eth.get_transaction_receipt(tx_hash)
    except Exception as e:
        return f"Error retrieving transaction receipt: {str(e)}"
    try:
        tx = client.eth.get_transaction(tx_hash)
    except Exception:
        tx = None
    details = f"Transaction Receipt for {tx_hash}:\n"
    details += f"  Status: {'Success' if receipt.status == 1 else 'Failure'}\n"
    details += f"  Block Number: {receipt.blockNumber}\n"
    details += f"  Gas Used: {receipt.gasUsed}\n"
    if tx is not None:
        details += f"  Gas Price: {Web3.from_wei(tx.gasPrice, 'gwei')} Gwei\n"
        tx_fee = tx.gasPrice * receipt.gasUsed
        details += f"  Transaction Fee: {Web3.from_wei(tx_fee, 'ether')} BNB\n"
    details += f"  Contract Address: {receipt.contractAddress}\n"
    details += f"  Logs Count: {len(receipt.logs)}\n"
    return details

class TransactionReceiptAction(BnbChainAction):
    name: str = "get_transaction_receipt"
    description: str = "Retrieve and display transaction details given a transaction hash."
    args_schema: type[BaseModel] = TransactionReceiptInput
    func = get_transaction_receipt

# --------------------------------------------------------------------
# 3. Recent Transactions Fetcher
# --------------------------------------------------------------------

class RecentTransactionsInput(BaseModel):
    address: str = Field(..., description="Address to fetch recent transactions for.")
    limit: int = Field(5, description="Number of transactions to retrieve.")

def get_recent_transactions(provider: BnbChainProvider, address: str, limit: int = 5) -> str:
    api_key = os.environ.get("BSCSCAN_API_KEY")
    if not api_key:
        return "BSCSCAN_API_KEY not set in environment."
    url = (
        f"https://api.bscscan.com/api?module=account&action=txlist&address={address}"
        f"&startblock=0&endblock=99999999&page=1&offset={limit}&sort=desc&apikey={api_key}"
    )
    try:
        response = requests.get(url, timeout=10)
        data = response.json()
        if data.get("status") != "1":
            return f"Error fetching transactions: {data.get('message', 'Unknown error')}"
        txs = data.get("result", [])
        result = f"Recent transactions for address {address}:\n"
        for i, tx in enumerate(txs, 1):
            value_bnb = Web3.from_wei(int(tx['value']), 'ether')
            status = "Success" if tx.get("isError") == "0" else "Failure"
            result += f"{i}. TxHash: {tx['hash']} – Block: {tx['blockNumber']} – Value: {value_bnb} BNB – Status: {status}\n"
        return result
    except Exception as e:
        return f"Error fetching transactions: {str(e)}"

class RecentTransactionsAction(BnbChainAction):
    name: str = "recent_transactions"
    description: str = "Retrieve the recent transactions for a given address."
    args_schema: type[BaseModel] = RecentTransactionsInput
    func = get_recent_transactions

# --------------------------------------------------------------------
# 4. Pending Transaction Tracker
# --------------------------------------------------------------------

class PendingTransactionsInput(BaseModel):
    address: str = Field(..., description="Address to track pending transactions for.")
    limit: int = Field(5, description="Maximum number of pending transactions to return.")

def get_pending_transactions(provider: BnbChainProvider, address: str, limit: int = 5) -> str:
    client = provider.get_current_client()
    try:
        pending_filter = client.eth.filter("pending")
        pending_tx_hashes = pending_filter.get_new_entries()
    except Exception as e:
        return f"Error creating pending filter: {str(e)}"
    
    matching_txs = []
    for tx_hash in pending_tx_hashes:
        try:
            tx = client.eth.get_transaction(tx_hash)
            if tx is None:
                continue
            if tx['from'].lower() == address.lower() or (tx.get('to') and tx['to'].lower() == address.lower()):
                matching_txs.append(tx)
                if len(matching_txs) >= limit:
                    break
        except Exception:
            continue
    if not matching_txs:
        return f"No pending transactions found for address {address}."
    result = f"Pending transactions for address {address}:\n"
    for i, tx in enumerate(matching_txs, 1):
        gas_price_gwei = Web3.from_wei(tx.gasPrice, 'gwei')
        result += f"{i}. TxHash: {tx.hash.hex()} – Gas Price: {gas_price_gwei} Gwei\n"
    return result

class PendingTransactionTrackerAction(BnbChainAction):
    name: str = "pending_transactions"
    description: str = "Track and list pending transactions for a given address."
    args_schema: type[BaseModel] = PendingTransactionsInput
    func = get_pending_transactions

# --------------------------------------------------------------------
# 5. Gas Price Trend Analyzer
# --------------------------------------------------------------------

class GasPriceTrendInput(BaseModel):
    num_blocks: int = Field(10, description="Number of recent blocks to analyze for gas price trend.")

def get_gas_price_trend(provider: BnbChainProvider, num_blocks: int = 10) -> str:
    client = provider.get_current_client()
    latest_block = client.eth.get_block('latest', full_transactions=True)
    latest_block_number = latest_block.number
    total_gas_price = 0
    tx_count = 0
    for i in range(latest_block_number, latest_block_number - num_blocks, -1):
        try:
            block = client.eth.get_block(i, full_transactions=True)
            for tx in block.transactions:
                total_gas_price += tx.gasPrice
                tx_count += 1
        except Exception:
            continue
    if tx_count == 0:
        return "No transactions found in recent blocks."
    avg_gas_price = total_gas_price // tx_count
    avg_gas_price_gwei = Web3.from_wei(avg_gas_price, 'gwei')
    return f"Average gas price over the last {num_blocks} blocks is approximately {avg_gas_price_gwei} Gwei."

class GasPriceTrendAnalyzerAction(BnbChainAction):
    name: str = "gas_price_trend"
    description: str = "Analyze recent blocks to determine the average gas price trend."
    args_schema: type[BaseModel] = GasPriceTrendInput
    func = get_gas_price_trend

# --------------------------------------------------------------------
# 6. Transaction Fee Estimator
# --------------------------------------------------------------------

class TransactionFeeEstimatorInput(BaseModel):
    gas_limit: int = Field(..., description="Gas limit for the transaction.")
    gas_price_gwei: int = Field(..., description="Gas price in Gwei.")

def estimate_transaction_fee(provider: BnbChainProvider, gas_limit: int, gas_price_gwei: int) -> str:
    gas_price_wei = Web3.to_wei(gas_price_gwei, 'gwei')
    fee_wei = gas_limit * gas_price_wei
    fee_bnb = Web3.from_wei(fee_wei, 'ether')
    return f"Estimated transaction fee: {fee_bnb} BNB (Gas Limit: {gas_limit}, Gas Price: {gas_price_gwei} Gwei)"

class TransactionFeeEstimatorAction(BnbChainAction):
    name: str = "transaction_fee_estimator"
    description: str = "Estimate the transaction fee based on provided gas limit and gas price."
    args_schema: type[BaseModel] = TransactionFeeEstimatorInput
    func = estimate_transaction_fee

# --------------------------------------------------------------------
# 7. Transaction Volume Analyzer
# --------------------------------------------------------------------

class TransactionVolumeAnalyzerInput(BaseModel):
    address: str = Field(..., description="Address to analyze transaction volume for.")
    hours: int = Field(24, description="Time period in hours to consider.")

def analyze_transaction_volume(provider: BnbChainProvider, address: str, hours: int = 24) -> str:
    api_key = os.environ.get("BSCSCAN_API_KEY")
    if not api_key:
        return "BSCSCAN_API_KEY not set in environment."
    
    # Get the current block number from the provider.
    client = provider.get_current_client()
    latest_block = client.eth.get_block('latest')
    latest_block_number = latest_block.number

    # Estimate the number of blocks produced in the given time period.
    # BSC's average block time is approximately 3 seconds.
    avg_block_time = 3  # seconds per block
    blocks_in_period = int((hours * 3600) / avg_block_time)
    startblock = latest_block_number - blocks_in_period
    if startblock < 0:
        startblock = 0

    # Use startblock and endblock for the API query.
    url = (
        f"https://api.bscscan.com/api?module=account&action=txlist&address={address}"
        f"&startblock={startblock}&endblock={latest_block_number}&sort=desc&apikey={api_key}"
    )
    try:
        response = requests.get(url, timeout=10)
        data = response.json()
        if data.get("status") != "1":
            return f"Error fetching transactions: {data.get('message', 'Unknown error')}"
        txs = data.get("result", [])
        total_volume = sum(int(tx['value']) for tx in txs)
        total_volume_bnb = Web3.from_wei(total_volume, 'ether')
        count = len(txs)
        avg_volume = total_volume_bnb / count if count > 0 else 0
        return (
            f"In the last {hours} hours (from block {startblock} to {latest_block_number}), address {address} "
            f"has a total transaction volume of {total_volume_bnb} BNB over {count} transactions, averaging "
            f"{avg_volume} BNB per transaction."
        )
    except Exception as e:
        return f"Error analyzing transaction volume: {str(e)}"

class TransactionVolumeAnalyzerAction(BnbChainAction):
    name: str = "transaction_volume_analyzer"
    description: str = "Analyze the total transaction volume for an address over a specified time period."
    args_schema: type[BaseModel] = TransactionVolumeAnalyzerInput
    func = analyze_transaction_volume

# --------------------------------------------------------------------
# 8. Token Transfer History
# --------------------------------------------------------------------

class TokenTransferHistoryInput(BaseModel):
    address: str = Field(..., description="Address to get token transfer history for.")
    token: str = Field("BNB", description="Token symbol or contract address. Use 'BNB' for native token.")
    limit: int = Field(5, description="Number of transfers to retrieve.")

def get_token_transfer_history(provider: BnbChainProvider, address: str, token: str = "BNB", limit: int = 5) -> str:
    api_key = os.environ.get("BSCSCAN_API_KEY")
    if not api_key:
        return "BSCSCAN_API_KEY not set in environment."
    
    if token.upper() == "BNB":
        url = (
            f"https://api.bscscan.com/api?module=account&action=txlist&address={address}"
            f"&startblock=0&endblock=99999999&page=1&offset={limit}&sort=desc&apikey={api_key}"
        )
    else:
        # Map token symbol to contract address if available
        contract_address = TOKEN_CONTRACT_ADDRESSES.get(token.upper(), token)
        url = (
            f"https://api.bscscan.com/api?module=account&action=tokentx&address={address}"
            f"&contractaddress={contract_address}&startblock=0&endblock=99999999&page=1&offset={limit}&sort=desc&apikey={api_key}"
        )
    try:
        response = requests.get(url, timeout=10)
        data = response.json()
        if data.get("status") != "1":
            return f"Error fetching transfer history: {data.get('message', 'Unknown error')}"
        txs = data.get("result", [])
        result = f"Token Transfer History for {token.upper()} on address {address}:\n"
        for i, tx in enumerate(txs, 1):
            value_bnb = Web3.from_wei(int(tx['value']), 'ether')
            result += f"{i}. TxHash: {tx['hash']} transferred {value_bnb} {token.upper()} at Block {tx['blockNumber']}\n"
        return result
    except Exception as e:
        return f"Error fetching token transfer history: {str(e)}"
    
class TokenTransferHistoryAction(BnbChainAction):
    name: str = "token_transfer_history"
    description: str = "Retrieve the history of token transfers for a given address and token."
    args_schema: type[BaseModel] = TokenTransferHistoryInput
    func = get_token_transfer_history



import asyncio
import logging
import readline  # noqa: F401
import sys

from dotenv import load_dotenv
from langchain_core.messages import HumanMessage
from langchain_openai import ChatOpenAI
from langgraph.checkpoint.memory import MemorySaver
from langgraph.prebuilt import create_react_agent

from bnb_chain_agentkit.agent_toolkits import BnbChainToolkit
from bnb_chain_agentkit.utils import BnbChainAPIWrapper

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
logger = logging.getLogger('Chatbot')

console_handler = logging.StreamHandler()
logger.addHandler(console_handler)

load_dotenv()


def initialize_agent():
    """Initialize the agent with BNB Chain."""
    # Initialize LLM.
    llm = ChatOpenAI(model='gpt-4o')

    # Configure BNB Chain Langchain Extension.
    bnb_chain = BnbChainAPIWrapper()

    # Initialize BNB Chain Toolkit and get tools.
    bnb_chain_toolkit = BnbChainToolkit.from_bnb_chain_api_wrapper(bnb_chain)
    tools = bnb_chain_toolkit.get_tools()
    print('Supported tools:')
    for tool in tools:
        print(tool.name)

    # Store buffered conversation history in memory.
    memory = MemorySaver()
    config = {'configurable': {'thread_id': 'BNB Chain Chatbot Example!'}}

    # Create React Agent using the LLM and BNB Chain tools.
    return create_react_agent(
        llm,
        tools=tools,
        checkpointer=memory,
        state_modifier="You are a helpful agent that can interact onchain using the BNB Chain toolkit. You are empowered to interact onchain using your tools. If you ever need funds, you can provide your wallet details and request funds from the user. If someone asks you to do something you can't do with your currently available tools, you must say so, and encourage them to implement it themselves using the BNB Chain. Be concise and helpful with your responses. Refrain from restating your tools' descriptions unless it is explicitly requested.",
    ), config


async def run_chat_mode(agent_executor, config):
    """Run the agent interactively based on user input."""
    print("Starting chat mode... Type 'exit' to end.")
    while True:
        try:
            user_input = input('\nUser: ')
            if user_input.lower() == 'exit':
                break

            # Run agent with the user's input in chat mode
            async for event in agent_executor.astream(
                {'messages': [HumanMessage(content=user_input)]},
                config,
                stream_mode='values',
            ):
                event['messages'][-1].pretty_print()

        except KeyboardInterrupt:
            print('Goodbye Agent!')
            sys.exit(0)
        except EOFError:  # Catch EOFError for handling Ctrl+D
            print('Goodbye Agent!')
            sys.exit(0)


async def main():
    """Start the chatbot agent."""
    agent_executor, config = initialize_agent()
    await run_chat_mode(agent_executor=agent_executor, config=config)


if __name__ == '__main__':
    print('Starting Agent...')
    asyncio.run(main())



import requests
from pydantic import BaseModel, Field
from bnb_chain_agentkit.actions.bnb_chain_action import BnbChainAction
from bnb_chain_agentkit.provider.bnb_chain_provider import BnbChainProvider

# Mapping from common token symbols to Coingecko IDs.
TOKEN_CG_IDS = {
    "BNB": "binancecoin",
    "BUSD": "binance-usd",
    "BTC": "bitcoin",
    "ETH": "ethereum",
    "USDC": "usd-coin",
    "DAI": "dai",
    "USDT": "tether",
}

class TokenPriceQueryInput(BaseModel):
    token: str = Field(..., description="Token symbol (e.g., BNB, BUSD, BTC, ETH) or a Coingecko token id.")
    vs_currency: str = Field("usd", description="The fiat currency to compare against (default 'usd').")

def token_price_query(provider: BnbChainProvider, token: str, vs_currency: str = "usd") -> str:
    token_id = TOKEN_CG_IDS.get(token.upper(), token.lower())
    url = f"https://api.coingecko.com/api/v3/simple/price?ids={token_id}&vs_currencies={vs_currency}"
    try:
        response = requests.get(url, timeout=10)
        if response.status_code != 200:
            return f"Error fetching price: HTTP {response.status_code}"
        data = response.json()
        if token_id not in data:
            return f"Token '{token}' not found in Coingecko API."
        price = data[token_id].get(vs_currency)
        return f"The current price of {token.upper()} is {price} {vs_currency.upper()}."
    except Exception as e:
        return f"Error fetching token price: {str(e)}"

class TokenPriceQueryAction(BnbChainAction):
    name: str = "token_price_query"
    description: str = "Fetch the current price of a token using the Coingecko API."
    args_schema: type[BaseModel] = TokenPriceQueryInput
    func = token_price_query



"""**Utilities** are the integration wrappers that LangChain uses to interact with third-party systems and packages."""

from bnb_chain_agentkit.utils.bnb_chain_api_wrapper import BnbChainAPIWrapper

__all__ = ['BnbChainAPIWrapper']



from typing import Union
import decimal
from decimal import (
    localcontext,
)


MIN_WEI = 0
MAX_WEI = 2**256 - 1


def format_units(number: int, decimals: int) -> Union[int, decimal.Decimal]:
    """
    Takes a number of wei and converts it to any other unit.
    """
    if number == 0:
        return 0

    if number < MIN_WEI or number > MAX_WEI:
        raise ValueError('value must be between 0 and 2**256 - 1')

    unit_value = decimal.Decimal(10) ** decimals

    with localcontext() as ctx:
        ctx.prec = 999
        d_number = decimal.Decimal(value=number, context=ctx)
        result_value = d_number / unit_value

    return result_value


def parse_units(number: Union[int, float, str, decimal.Decimal], decimals: int) -> int:
    """
    Takes a number of a unit and converts it to wei.
    """
    if decimals < 1:
        raise ValueError('Decimals must be greater than 0')

    if isinstance(number, int) or isinstance(number, str):
        d_number = decimal.Decimal(value=number)
    elif isinstance(number, float):
        d_number = decimal.Decimal(value=str(number))
    elif isinstance(number, decimal.Decimal):
        d_number = number
    else:
        raise TypeError('Unsupported type. Must be one of integer, float, or string')

    s_number = str(number)
    unit_value = decimal.Decimal(10) ** decimals

    if d_number == decimal.Decimal(0):
        return 0

    if d_number < 1 and '.' in s_number:
        with decimal.localcontext() as ctx:
            multiplier = len(s_number) - s_number.index('.') - 1
            ctx.prec = multiplier
            d_number = decimal.Decimal(value=number, context=ctx) * 10**multiplier
        unit_value /= 10**multiplier

    with decimal.localcontext() as ctx:
        ctx.prec = 999
        result_value = decimal.Decimal(value=d_number, context=ctx) * unit_value

    if result_value < MIN_WEI or result_value > MAX_WEI:
        raise ValueError('Resulting wei value must be between 0 and 2**256 - 1')

    return int(result_value)



"""BNB Chain Toolkit."""

from langchain_core.tools import BaseTool
from langchain_core.tools.base import BaseToolkit

from bnb_chain_agentkit.actions import BNB_CHAIN_ACTIONS
from bnb_chain_agentkit.tools import BnbChainTool
from bnb_chain_agentkit.utils import BnbChainAPIWrapper


class BnbChainToolkit(BaseToolkit):
    """BNB Chain Toolkit.

    *Security Note*: This toolkit contains tools that can read and modify
        the state of a service; e.g., by creating, deleting, or updating,
        reading underlying data.

        For example, this toolkit can be used to create wallets, transactions,
        and smart contract invocations on BNB Chain.

    Setup:
        See detailed installation instructions here:
        https://python.langchain.com/docs/integrations/tools/bnb_chain/#installation

        You will need to set the following environment
        variables:

        .. code-block:: bash

            export PRIVATE_KEY="private-key"
            export PROVIDER_URL="provider-url"

    Instantiate:
        .. code-block:: python

            from bnb_chain.agent_toolkits import BnbChainToolkit
            from bnb_chain.utils import BnbChainAPIWrapper

            bnb_chain = BnbChainAPIWrapper()
            bnb_chain_toolkit = BnbChainToolkit.from_bnb_chain_api_wrapper(bnb_chain)

    Tools:
        .. code-block:: python

            tools = bnb_chain_toolkit.get_tools()
            for tool in tools:
                print(tool.name)

        .. code-block:: none

            get_balance
            transfer
    Use within an agent:
        .. code-block:: python

            from langchain_openai import ChatOpenAI
            from langgraph.prebuilt import create_react_agent

            # Select example tool
            tools = [tool for tool in toolkit.get_tools() if tool.name == "get_balance"]
            assert len(tools) == 1

            llm = ChatOpenAI(model="gpt-4o-mini")
            agent_executor = create_react_agent(llm, tools)

            example_query = "Check my BNB balance"

            events = agent_executor.stream(
                {"messages": [("user", example_query)]},
                stream_mode="values",
            )
            for event in events:
                event["messages"][-1].pretty_print()

        .. code-block:: none

            ================================ Human Message =================================

            Check my BNB balance
            ================================== Ai Message ==================================
            Tool Calls:
            get_balance (call_w4HfhOLX9d5lH7emf1QKQCF4)
            Call ID: call_w4HfhOLX9d5lH7emf1QKQCF4
            Args:
                account: None
                token: None
            ================================= Tool Message =================================
            Name: get_balance

            Balances for account 0x1234 of BNB:
            6145615640500000000 (decimals: 18)
            ================================== Ai Message ==================================

            Your BNB balance is 6.1456156405 BNB.

    Parameters
    ----------
        tools: List[BaseTool]. The tools in the toolkit. Default is an empty list.

    """

    tools: list[BaseTool] = []

    @classmethod
    def from_bnb_chain_api_wrapper(cls, bnb_chain_api_wrapper: BnbChainAPIWrapper) -> 'BnbChainToolkit':
        """Create a BnbChainToolkit from a BnbChainAPIWrapper.

        Args:
            bnb_chain_api_wrapper: BnbChainAPIWrapper. The BNB Chain API wrapper.

        Returns:
            BnbChainToolkit. The BNB Chain toolkit.

        """
        actions = BNB_CHAIN_ACTIONS

        tools = [
            BnbChainTool(
                name=action.name,
                description=action.description,
                bnb_chain_api_wrapper=bnb_chain_api_wrapper,
                args_schema=action.args_schema,
                func=action.func,
                async_func=action.async_func,
            )
            for action in actions
        ]

        return cls(tools=tools)  # type: ignore[arg-type]

    def get_tools(self) -> list[BaseTool]:
        """Get the tools in the toolkit."""
        return self.tools



import logging
from collections.abc import Callable
from typing import Optional

from pydantic import BaseModel, Field
from web3 import Web3

from bnb_chain_agentkit.actions.abi import ERC20_ABI
from bnb_chain_agentkit.actions.bnb_chain_action import BnbChainAction
from bnb_chain_agentkit.actions.utils.units import parse_units
from bnb_chain_agentkit.provider.bnb_chain_provider import BnbChainProvider

logger = logging.getLogger(__name__)

TRANSFER_PROMPT = """
This tool will transfer an asset from the wallet to another onchain address.

It takes the amount, token, and recipient as inputs.

Important notes:
- The token could be a valid token address or "BNB".
- Ensure sufficient balance of the input asset before transferring
- When sending native assets (e.g. 'BNB' on BSC mainnet), ensure there is sufficient balance for the transfer itself AND the gas cost of this transfer
"""


class TransferInput(BaseModel):
    """Input argument schema for transfer action."""

    amount: str = Field(description='The amount of the asset to transfer, e.g. `15`, `0.000001`')
    token: str = Field(
        'BNB',
        description='The token to transfer, e.g. `BNB` or `0x...`.',
    )
    recipient: str = Field(
        description='The recipient to transfer the funds, e.g. `0x58dBecc0894Ab4C24F98a0e684c989eD07e4e027`',
    )


def transfer(
    provider: BnbChainProvider,
    amount: str,
    token: str,
    recipient: str,
) -> str:
    """Transfer a specified amount of an asset to a recipient onchain.

    Args:
        provider (BnbChainProvider): The provider to use for the transfer.
        amount (str): The amount of the asset to transfer, e.g. `15`, `0.000001`.
        token (str): The token to transfer, e.g. `BNB` or `0x...`.
        recipient (str): The recipient to transfer the funds (e.g. `0x58dBecc0894Ab4C24F98a0e684c989eD07e4e027`).

    Returns:
        str: A message containing the transfer details.
    """

    client = provider.get_current_client()
    recipient = Web3.to_checksum_address(recipient)

    if token == 'BNB':
        tx_hash = client.eth.send_transaction(
            {
                'to': recipient,
                'value': Web3.to_wei(amount, 'ether'),
            }
        )
        token_name = token
    else:
        token_address = Web3.to_checksum_address(token)
        contract = client.eth.contract(token_address, abi=ERC20_ABI)

        token_name = contract.functions.name().call()
        decimals = contract.functions.decimals().call()
        amount_wei = parse_units(amount, decimals)
        tx_hash = contract.functions.transfer(recipient, amount_wei).transact()

    client.eth.wait_for_transaction_receipt(tx_hash)
    return (
        f'Transferred {amount} {token_name} to {recipient}.\nTransaction hash for the transfer: {tx_hash.to_0x_hex()}'
    )


class TransferAction(BnbChainAction):
    """Transfer action."""

    name: str = 'transfer'
    description: str = TRANSFER_PROMPT
    args_schema: Optional[type[BaseModel]] = TransferInput
    func: Optional[Callable[..., str]] = transfer
